/*
   $Id: crawextractjpeg.c,v 1.7 2005/03/24 06:17:16 dmg Exp $

   Extract the embedded jpeg from a Canon(TM) raw photograph.


 * Copyright (C) 2004 Daniel M. German
 *
 * Author: Daniel M. German dmg _at_ turingmachine _dot_ org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>

#include <getopt.h>

#include <dirent.h>
#include <sys/stat.h>
#include <libgen.h>
#include <string.h>

#include <assert.h>

#include "libcraw.h"

/* Make sure we use GETOPT */

#define GETOPT


/* For some reason in linux this is already defined, but not in OS X*/
#ifndef PATH_MAX
#include <sys/syslimits.h>
#endif


#define USAGE "JPEP extractor Version " VERSION "\n\
Extracts the embedded JPEG from a Canon(TM) Raw Photograph\n\
Copyright (C) 2004 Daniel M. German\n\
Author: Daniel M. German dmgerman at uvic dot ca\n\
\n\
Usage: %s [OPTION] filename ...\n\
\n\
 Options\n\
  -v --verbose           increase verbosity\n\
  -o --overwrite         overwrite output file\n\
  -d --outputDir         directory where output files should\n\
                         be placed\n\
  -c --createdir         create output directory if it does\n\
                         not exits\n\
\n\
         \n\
By default the output files are created in the current directory. \n\
\n\
The source filename is stripped of the .crw extension (case \n\
insensitive and replaced by .jpg.\n"



/* Flag set by `--verbose'. */
static int verbose = 0;
static int overwrite = 0;
static int createDirectory = 0;

static char copyBuffer[20480]; /* 20k is not onerous these days */

static int Process_File(char *fileName, char *outputFileName, int overwrite)
{
   FILE *input;
   FILE *output;
   int returnValue = 0;
   long toRead, bytesRead, bytesWritten;

   unsigned long dataOffset;
   unsigned long dataLength;

   craw_photo_type *photo;
   
   /* Let us first check if the source file exists */
   if ((input = fopen(fileName, "rb")) == NULL) {
      fprintf(stderr, "\n   Input file %s does not exist\n", fileName);
      returnValue = 0;
      goto justreturn;
   }

   if (!overwrite) {
      /* If we can't overwrite, we first check if the file exists*/
      if ((output = fopen(outputFileName, "rb")) != NULL) {
	 fprintf(stderr, "\n   Output file %s exists. I will not overwrite\n", outputFileName);
	 returnValue = 0;
	 goto closeinout;
      } 
   }

   if ((output = fopen(outputFileName, "wb")) == NULL) {
      fprintf(stderr, "\n   Unable to create output file %s\n", outputFileName);
	 returnValue = 0;
	 goto closein;
   } 

   /* So now we need to do the conversion */
   
   /* BOth files are open and ready to go */
   
   if ((photo = craw_Load_Photo(input)) == NULL) {
      fprintf(stderr,"\n   File %s does not appear to be a Canon photo\n", fileName);
      returnValue = 0;
      goto closeinout;
   }

   /*
   if (craw_Find_Jpeg_Location_Only(photo *pPhoto, long dataOffset, 
				    long lengthData);

   */
   
   if (!craw_Get_Jpeg_Location(photo, &dataOffset, &dataLength)) {
      fprintf(stderr,"\n   File %s does not contain an embedded JPEG\n", fileName);
      returnValue = 0;
      goto destroyphoto;
   }

   /*
   printf("We need to copy from %ld %ld bytes\n", dataOffset, dataLength);
   */
   /* First seek the input file */
   if (fseek(input, dataOffset, SEEK_SET) != 0) {
      fprintf(stderr,"\n   File %s does not contain an embedded JPEG,"
	      "but the header said so, please report to maintainers\n", fileName);
      returnValue = 0;
      goto destroyphoto;
   }
   /* REady to go */
   toRead = dataLength;
   while (toRead > 0) {
      bytesRead = fread(copyBuffer, 1, sizeof(copyBuffer), input);
      if (bytesRead == 0) {
	 fprintf(stderr,"\n  File %s does not contain an complete embedded JPEG,"
		 "but the header said so, please report to maintainers\n", fileName);
	 returnValue = 0;
	 goto destroyphoto;
      }
      toRead -= bytesRead;
      bytesWritten = fwrite(copyBuffer, 1, bytesRead, output);
      assert(bytesWritten == bytesRead);
      
   }
   /* A lot of the people reading this code will frown upon my use of
      gotos, but mind you, this is the way a lot of real code is written, 
      it is C way to do exception handling, trust me :) */
   
   returnValue = 1;


 destroyphoto:
   craw_Destroy_Photo(photo);
 closeinout:
   fclose(output);
 closein:
   fclose(input);
 justreturn:
   return returnValue;

}


int main (argc, argv)
     int argc;
     char **argv;
{

   char *directoryName = NULL;

#ifndef GETOPT

   int optind = 1;

#else

   int keepParsingOptions = 1;
   int c;

   while (keepParsingOptions) {
      static struct option longOptions[] = {
	 /* These options set a flag. */
	 {"verbose",    no_argument,       &verbose,   1},
	 {"overwrite",  no_argument,    &overwrite, 1},
	 {"createdir",  no_argument,    &createDirectory, 1},
	 /* These options don't set a flag.

	    We distinguish them by their indices. */
	 {"outputDir",  required_argument, 0, 'd'},

	 {0, 0, 0, 0}
      };
      /* getopt_long stores the option index here. */
      int optionIndex = 0;
      
      c = getopt_long (argc, argv, "vocd:",
                       longOptions, &optionIndex);
      

      if (c == -1) {
	 /* Detect the end of the options. */
	 keepParsingOptions = 0;
      } else {
	 switch (c) {
	 case 'o':
	    overwrite = 1;
	 case 'v':
	    verbose = 1;
	 case 'c':
	    createDirectory = 1;
	 case 'd':
	    directoryName = optarg;
	    break;
	 case '?':
	    fprintf(stderr,  USAGE, argv[0]);
	    exit(1);
	 }
      }
   }
#endif
   /* Print any remaining command line arguments (not options). */

   if (directoryName != NULL) {
      /* Make sure the directory exists */
      DIR *dir;
      if ((dir=opendir (directoryName)) != NULL) {
	 closedir(dir);
      } else {
	 if (!createDirectory) {
	    fprintf(stderr, "Unable to access directory %s: ", directoryName);
	    perror("");
	    exit(1);
	 } else {
	    /* We have to create the directory */
	    mkdir (directoryName, S_IRWXU);
	 }
      }
   }

   if (optind >= argc) {
      fprintf(stderr,  USAGE, argv[0]);
      exit(1);
   }

   while (optind < argc) {
      char *path = argv[optind++];
      char *filename;
      char newFileName[PATH_MAX+1];
      
      filename = basename (path);
      
      
      if (directoryName != NULL) {
	 /* Make sure we copy only as many bytes as we space,
	    we don't want a buffer overrun */
	 
	 strncpy(newFileName, directoryName, PATH_MAX);
	 strcat(newFileName, "/");
	 strncat(newFileName, filename, PATH_MAX - strlen(newFileName));
      } else {
	 strncpy(newFileName, filename, PATH_MAX);
      }
      
      /* Replace Extension */
      
      if (strlen(filename) >= 4 &&
	  strcasecmp(&filename[strlen(filename) - 4], ".crw") == 0)
	 strcpy(&newFileName[strlen(newFileName) - 4], ".jpg");
      else {
	 if (strlen(newFileName) > PATH_MAX - 4) {
	    fprintf(stderr, "Illegal file name %s\n", path);
	    continue;
	 }
	 strcat(newFileName, ".jpg");
      }
      
      if (verbose) {
	 fprintf(stderr, "Processing file %s -> %s...", path, newFileName);
      }
      /* We are ready to go */
      
      Process_File(path, newFileName, overwrite);
   }
   return (0);
}


