/*
   $Id: libciff++.cpp,v 1.5 2005/03/24 06:12:15 dmg Exp $

   Library to read the records from a RAW file from Canon

   Credits:

     David Coffin:  He reversed engineered the format
     David Burren:  Embedded exif record
     Peter Galbavy: His documentation of the format helped a lot


 * Copyright (C) 2004 Daniel M. German
 *
 * Author: Daniel M. German dmg _at_ turingmachine _dot_ org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


/* This code implements a generic class that parses the CIFF data,
   callbacks are empty shells that can be defined in the subclass and
   called by way of polymorphism */


#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>


#include "libraw.h"
#include "crawdebug.h"
#include "libciff.h"

/* Masks */
#define CIFF_TYPE_MASK             0xc000  /* 1100 0000 0000 0000 */
#define CIFF_TYPE_IN_HEAP_SPACE    0x0000
#define CIFF_TYPE_IN_RECORD_ENTRY  0x1000


#define RECORD_RAW               0x2005
#define RECORD_JPEG              0x2007
#define RECORD_CAMERA_HEAP      (0x2807)
#define RECORD_IMAGE_DETAILS_HEAP      (0x2804)
#define RECORD_CAMERA_SETUP_HEAP       (0x300b)

#define RECORD_MEASURED_DATA_HEAP     0x300a


#define CIFF_MASK_STRING         0x0800   /* 0000 1000 */
#define CIFF_MASK_TYPE_RECORD    0x1000   /* 0001 0000 */
#define CIFF_TYPE_8_BYTES        0x5000   /* 1001 0000 */

#define RECORD_SENSOR_PROPERTIES (CIFF_MASK_TYPE_RECORD | 0x0031)
#define RECORD_EXPOSURE_INFO     (CIFF_MASK_TYPE_RECORD | 0x002a)
#define RECORD_CUSTOM_FUNCTIONS  (CIFF_MASK_TYPE_RECORD | 0x0033)  /* Bunch of byte values */



#define RECORD_IMAGE_DETAILS     (CIFF_MASK_TYPE_RECORD | 0x0810)  /* INcludes rotarion information */
#define RECORD_IMAGE_DATE        (CIFF_MASK_TYPE_RECORD | 0x080e)

/* ---------------------------------------------------------------------- */

#define RECORD_FILENAME          (CIFF_MASK_STRING | 0x0016)
#define RECORD_THM_FILENAME      (CIFF_MASK_STRING | 0x0017)
#define RECORD_OWNER_NAME        (CIFF_MASK_STRING | 0x0010)

/* This appers to be a user filled field with the description of the photo*/
#define RECORD_IMAGE_DESCRIPTION (CIFF_MASK_STRING | 0x0005)
#define RECORD_IMAGE_NAME        (CIFF_MASK_STRING | 0x0015) /* */

#define RECORD_CAMERA_NAME       (CIFF_MASK_STRING | 0x000a)
#define RECORD_CAMERA_FIRMWARE   (CIFF_MASK_STRING | 0x000b)


#define RECORD_FOCAL_LENGTH       (CIFF_TYPE_8_BYTES | 0x0029)

#define RECORD_IMAGE_FORMAT       (0x5800 | 0x0003)
#define RECORD_TARGET_IMAGE_TYPE  (CIFF_TYPE_8_BYTES | 0x000a) /* 8 Bytes, Image description? */
#define RECORD_UNKNOWN3           (0x5800 | 0x0004)
#define RECORD_UNKNOWN4           (0x3000 | 0x0003)
#define RECORD_UNKNOWN5           (0x5800 | 0x0014)
#define RECORD_CAMERA_SPECIFICATION_HEAP           (0x3000 | 0x0004)
#define RECORD_BODY_SENSITIVITY   (0x5000 | 0x001c)
#define RECORD_ROM_OP_MODE        (0x4800 | 0x000d)
#define RECORD_UNKNOWN9           (0x0000 | 0x0036)
#define RECORD_UNKNOWN10          (0x5800 | 0x0034)
#define RECORD_UNKNOWN11          (0x1000 | 0x00c0)
#define RECORD_UNKNOWN12          (0x1000 | 0x00c1)
#define RECORD_UNKNOWN13          (0x1000 | 0x00c2)
#define RECORD_UNKNOWN14          (0x1000 | 0x00aa)
#define RECORD_UNKNOWN15          (0x1000 | 0x00a8)
#define RECORD_UNKNOWN16          (0x1000 | 0x00a9)

#ifdef TOREMOVE
#define RECORD_UNKNOWN33           (CIFF_TYPE_8_BYTES | 0x0004) /* 8 Bytes */
#define RECORD_UNKNOWN44           (CIFF_TYPE_8_BYTES | 0x0003)
#endif

#define RECORD_DECODING_TABLE      (0x1800 | 0x0035)

#define RECORD_PHOTO_SERIALNUMBER (0x5800 | 0x0017) /* 8 Bytes */


#define RECORD_CAMERA_SERIALNUMBER      (0x5800 | 0x000b)

#define RECORD_PHOTO_INFO          (CIFF_MASK_TYPE_RECORD | 0x002d)

#define RECORD_WHITE_BALANCE      (CIFF_MASK_TYPE_RECORD | 0x00a9) /* Whitebalance Table */

#define RECORD_UNKNOWN1          (CIFF_TYPE_8_BYTES | 0x0028)

typedef struct {
  int recordType;
  char *name;
} ciff_record_name;

ciff_record_name recordNames[] = {
  {RECORD_RAW,	"Raw photo from CCD"},
  {RECORD_JPEG,	"Embedded JPEG"},

  {RECORD_EXPOSURE_INFO,	"Exposure information"},
  {RECORD_CUSTOM_FUNCTIONS,	"Custom FNs"},
  {RECORD_IMAGE_DETAILS,	"Image details"},
  {RECORD_IMAGE_DATE,	"Image date"},

  {RECORD_FILENAME,	"Filename"},
  {RECORD_THM_FILENAME,	"Thumbnail Filename"},
  {RECORD_OWNER_NAME,	"Owner name"},
  {RECORD_IMAGE_DESCRIPTION,	"Image description"},
  {RECORD_IMAGE_NAME,	"Image name"},
  {RECORD_CAMERA_NAME,	"Camera model name"},
  {RECORD_CAMERA_FIRMWARE,	"Camera firmware"},

  {RECORD_FOCAL_LENGTH,	"Focal length"},
  {RECORD_PHOTO_SERIALNUMBER,	"Photo serial number"},
  {RECORD_PHOTO_INFO    ,	"Photo information"},
  {RECORD_CAMERA_SERIALNUMBER,  "Camera serial number"},

  {RECORD_MEASURED_DATA_HEAP,   "Measured data heap"},

  {RECORD_WHITE_BALANCE,	"White balance table"},

  {RECORD_SENSOR_PROPERTIES,	"Sensor properties"},
  {RECORD_CAMERA_HEAP,          "Camera heap"},
  {RECORD_IMAGE_DETAILS_HEAP,   "Image details heap"},
  {RECORD_CAMERA_SETUP_HEAP,    "Camera setup heap"},

  {RECORD_DECODING_TABLE,        "Decoding table"},

  {RECORD_UNKNOWN1,            "ISO ???"},

  {RECORD_IMAGE_FORMAT,         "Image format"},
  {RECORD_TARGET_IMAGE_TYPE,    "Target Image Type"},
  {RECORD_UNKNOWN3,             "*Record ID?"},
  {RECORD_UNKNOWN4,             "*Measured Info (Heap)"},
  {RECORD_UNKNOWN5,             "*Measured EV?"},
  {RECORD_CAMERA_SPECIFICATION_HEAP,   "Camera Specification Heap"},
  {RECORD_BODY_SENSITIVITY,             "Body sensitivity"},
  {RECORD_ROM_OP_MODE,             "ROM Operation Mode"},
  {RECORD_UNKNOWN9,               "*Unknown 9"},
  {RECORD_UNKNOWN10,             "**Unknown 10"},
  {RECORD_UNKNOWN11,             "**Unknown 11"},
  {RECORD_UNKNOWN12,             "**Unknown 12"},
  {RECORD_UNKNOWN13,             "**Unknown 13"},
  {RECORD_UNKNOWN14,             "**Unknown 14"},
  {RECORD_UNKNOWN15,             "**Unknown 15"},
  {RECORD_UNKNOWN16,             "**Unknown 16"},


  {0xFFFF, NULL},
};

char* whiteBalanceNames[]  = {
  "auto",
  "daylight",
  "cloudy",
  "tungsten",
  "fluorescent",
  "flash",
  "preset",
};

char *focusingModesNames[] = {
  "One shot",
  "AI Servo",
  "Unknown focusing mode",
  "Manual",
};

char *meteringModesNames[] = {
  "Unknown Metering Mode 0",
  "Unknown Metering Mode 1",
  "Unknown Metering Mode 2",
  "Evaluative",
  "Partial",
  "Centre",
};

char *exposureModesNames[] = {
  "Easy Mode (see Easy mode shooting)",
  "Program",
  "TV",
  "AV",
  "Manual",
  "A-Dept",
};

char *fileFormatsNames[] = {
  "",
  "JPEG",
  "Canon CRW",
};

char *compressionTypesNames[] = {
  "lossy JPEG",
  "none",
  "non-quantization JPEG",
};


char *targetImageTypesNames[] = {
  "Real-world subject",
  "Written document",
};


char *macroModesNames[] = {
  "",
  "Macro",
  "Normal",
};

char *flashModesNames [] = {
  "flash not fired",
  "auto",
  "on",
  "red-eye reduction",
  "slow synchro",
  "auto + red-eye reduction",
  "on + red-eye reduction",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "",
  "external flash (not set on D30)",
};

char *imageSizesNames [] = {
  "Large",
  "Medium",
  "Small",
};


char *easyShootingModesNames [] = {
  "Full Auto",
  "None",
  "Landscape",
  "Fast Shutter",
  "Slow Shutter",
  "Night",
  "B&W",
  "Sepia",
  "Portrait",
  "Sports",
  "Macro / Close-Up",
  "Pan Focus",
};

char *captureModesNames [] = {
  "Unknown",
  "JPEG",
  "RAW",
  "Movie",
};


typedef struct {
  char *name;
  int  numberOptions;
  char *options[5]; /* if a camera comes with more options, we'll bump it up */
} custom_function_name_record;


custom_function_name_record customFunctionsNamesD60 [] = {
  {"NOT USED",0,"","","",""}, /* It starts from 0, so this record is always unused */
  {"",0,"","","",""},
  {"Shutter button/AE lock button",4,"AF/AE lock","AE lock/AF","AF/AF lock","AE+release/AE+AF"},
  {"Mirror Lockup",2,"Disable","Enable","",""},
  {"Tv/Av and exposure level",2,"1/2 stop","1/3 stop","",""},
  {"AF-assit beam/Flash firing",4,"Emits/Fires",
   "Does not emit/Fires","Only ext. flash emits/Fires","Emits/Does not fire"},

  {"Shutter speed in Av mode",2,"Auto","1/200 Fixed","",""},
  {"AEB sequence/auto cancellation",4,
   "0 -> - -> + / Enable",
   "0 -> - -> + / Disable",
   "- -> 0 -> + / Enable",
   "- -> 0 -> + / Disable"},
  {"Shutter curtain sync",2,"1st Curtain Sync","2nd Curtain Sync","",""},
  {"Lens AF stop button Fn. Switch",3,"AF Stop","AF Start","AE lock while metering",""},
  {"Auto reduction of fill flash",2, "Enable","Disable","",""},

  {"Menu button return position",3,"Top","Previous (top if turned off)","Previous",""},
  {"SET button func. when shooting",4,"Default","Change Quality","Change ISO speed","Change Parameters"},
  {"Sensor cleaning",2,"Disable","Enable","",""},
  {"Superimposed display",2,"On","Off","",""},
  {"Shutter release w/o CF card",2,"Possible","Not possible","",""},

  {"Unknown", 4, "", "", "", ""},
  {"Unknown", 4, "", "", "", ""},


};


typedef struct {
   int4           xSize;
   int4           ySize;
   char           unknown1[20];

  /* Some examples of this record
    Values as bytes
      x00 x0C x00 x00 x00 x08 x00 x00 x00 x00 x80 x3F x00 x00 x00 x00
      x08 x00 x00 x00 x18 x00 x00 x00 x01 x01 x00 x00
     Values as bytes decimal
	0  12   0   0   0   8   0   0   0   0 128  63   0   0   0   0
	8   0   0   0  24   0   0   0   1   1   0   0
     Values as int2
	3072      0   2048      0
	   0  16256      0      0
	   8      0     24      0
	 257      0
     Values as int4
	    3072       2048 1065353216          0
	       8         24        257
  */

} image_description_ciff;

#define CIFF_IMAGE_DESCRIPTION_RECORD_SPEC = "44c";


Ciff_Parser::Ciff_Parser()
{
  this->recursionLevel = 0;

  this->outputBytesPerRow = 16;
  this->outputInt4PerRow = 4;
  this->outputInt2PerRow = 4;

  this->pPhoto = NULL;
}

char *Ciff_Parser::Find_Record_Name(short type)
{
  int i=0;
  while (recordNames[i].name != NULL) {
    if (recordNames[i].recordType == type) {
      return recordNames[i].name;
    }
    i++;
  }
  return "**Unknown record";
}

void Ciff_Parser::Translate_Image_Format(ciff_translated_image_format *imageFormat,
					 void *data)
{
  assert(sizeof(*imageFormat) == 8);
  memcpy(imageFormat, data, 8);



  Fix_Endianism_In_Record(imageFormat, CIFF_IMAGE_FORMAT_RECORD_SPEC, 8, pPhoto->byteOrder);

  /* We need the Most Significant 16 bits in the format
     and the least significant 16 bits are the compression

  in a big endian they are already there.

  But in the littleendian we have to "swap"  them
   */

  /* Most significant 16 bits is fileformat */
  if (Host_Endianism() == little_endian) {
    int temp = imageFormat->compressionType;
    imageFormat->compressionType = imageFormat->fileFormat;
    imageFormat->fileFormat = temp;
  }


}

void Ciff_Parser::Translate_Custom_Functions(ciff_translated_custom_functions **pCustomFNdata,
					     void *data)
{
  /* This function will allocate the memory for the customFunctions record */

  u_int16_t entries=0;

  /* Get the number of entries: (size -4)/2 */

  memcpy(&entries, data, sizeof(u_int16_t));

  Fix_Endianism_In_Record(&entries, "2", 2, pPhoto->byteOrder);


  entries = (entries -4) /2;

  *pCustomFNdata = NULL;


  if ((*pCustomFNdata = (ciff_translated_custom_functions*)
       malloc(sizeof(ciff_custom_function_entry) *
			       entries + sizeof(int))) == NULL) {
    fprintf(stderr, "Not enough memory to allocate custom functions\n");
    exit(1);
  }
  


  (*pCustomFNdata)->numberEntries = entries;

  for(int i=0;i<entries;i++) {
    int value, index;
    char *ptr;
    ptr = (char*)data;
    index = *(ptr + i*2 + 5);

    value =  *(ptr + i*2 + 4);

    (*pCustomFNdata)->entries[i].index = index;
    (*pCustomFNdata)->entries[i].value = value;
  }
}

void Ciff_Parser::Translate_Photo_Serial_Number_Record(ciff_translated_photo_serial_number *serialNumber,
						       void *data)
{
  assert(sizeof(ciff_translated_photo_serial_number) == 8);


  memcpy(serialNumber, data, 8);
  Fix_Endianism_In_Record(serialNumber, CIFF_PHOTO_SERIAL_NUMBER_RECORD_SPEC, 8, pPhoto->byteOrder);


}


void Ciff_Parser::Translate_Exposure_Record(ciff_translated_photo_exposure_information *goodExposure,
					    ciff_photo_exposure_information *exposure)
{
  float temp;

  /* Copy values that are verbatim */
  switch (exposure->iso) {
  case 160:
    goodExposure->iso = 100;
    break;
  case 192:
    goodExposure->iso = 200;
    break;
  case 224:
    goodExposure->iso = 400;
    break;
  case 256:
    goodExposure->iso = 800;
    break;
  case 268:
    goodExposure->iso = 1000;
    break;
  case 288:
    goodExposure->iso = 1600;
    break;
  default:
    fprintf(stderr,  "We don't know how to deal with this ISO %d\n", exposure->iso);
    exit(1);
  }

  /* Translate av
     Av = 2 ^ (n/64)
  */


  goodExposure->targetAV = exp2(exposure->targetAV/64.0);

  /* Should we worry about division by zero?, we better */
  temp = exp2(exposure->targetTV/32.0);
  assert(temp != 0.0);
  goodExposure->targetTV = 1.0/temp;

  goodExposure->exposureCompensation = exposure->exposureCompensation;
  goodExposure->whiteBalance = exposure->whiteBalance;
  goodExposure-> autoExposureCompensation = exposure->autoExposureCompensation;
  goodExposure->flashExposureCompensation = exposure->flashExposureCompensation;

  goodExposure->autoExposureBracketingCompensation = exposure->autoExposureBracketingCompensation;

  /* Actual TV */
  goodExposure->actualAV = exp2(exposure->actualAV/64.0);

  temp = exp2(exposure->actualTV/32.0);
  assert(temp != 0.0);
  goodExposure->actualTV = 1.0/temp;



}





int Ciff_Parser::Act_On_Record(unsigned short type)
{
  /* THis is a bad idea, we need to read all of them, and act on some of them.
     currently if this returns false, the record is not read */

#ifdef TEST
  if (type ==  RECORD_IMAGE_FORMAT
      || type == RECORD_CUSTOM_FUNCTIONS
      || type == RECORD_IMAGE_DETAILS
      || type == RECORD_RAW
      || type == RECORD_JPEG
      || type == RECORD_MEASURED_DATA_HEAP
      || type == RECORD_TARGET_IMAGE_TYPE
      || type == RECORD_PHOTO_SERIALNUMBER
      || type == RECORD_FILENAME
      || type == RECORD_IMAGE_DATE
      )
    return 0;
#endif
  //  if ()
    return 1;

  return 0;
}


void *Ciff_Parser::Read_Data(unsigned int length, int offset)
{
  void *data = NULL;

  if (fseek(pPhoto->file, offset, SEEK_SET) != 0) {
    fprintf(stderr, "Unable to seek to data for record at offset %d\n", offset);
    exit(1);
    return NULL;
  }
  data = malloc(length);
  if (data == NULL) {
    fprintf(stderr, "Unable to allocate memory of size %d\n", length);
    exit(1);
    return NULL;
  }

  if (fread(data, 1, length, pPhoto->file) != length) {
    fprintf(stderr, "Unable to read data record of size %d\n", length);
    free(data);
    exit(1);
    return NULL;
  }
  return data;
}

void Ciff_Parser::Process_Record_Type_String(unsigned short type, char *str)
{
  printf( "%s    NULL terminated string: [%s]\n", outputPrefix, str);
}

void Ciff_Parser::Process_Record_Type_Date(unsigned short type,
					   ciff_date *date)
{

  printf( "%s    Date: %s\n"
	  "%s    Time Zone %d\n",
	  outputPrefix,
	  ctime((time_t*)&(date->dateTT)),
	  outputPrefix,
	  date->timeZoneCode
	  );
}

void Ciff_Parser::Print_Bytes_Hex(int length, void *data)
{
  int i=0;

  while (i<length) {
    int j;
    printf("%s    ", outputPrefix);
    for (j=0;j <outputBytesPerRow && i < length;j++) {
      unsigned char value = ((unsigned char*)data)[i];
      printf("x%02X ", value);
      i++;
    }
    printf("\n");
  }
}

void Ciff_Parser::Process_Record_Custom_Functions(unsigned short type,
						  ciff_translated_custom_functions *data)
{

  if (strcmp(pPhoto->cameraName, "Canon EOS D60") ==0) {
    for (int i=0; i< data->numberEntries; i++) {
      printf("%s    Index %2d Entry %2d Value %2d FN %-30s Value: %-20s\n", outputPrefix,
	     i,
	     data->entries[i].index,
	     data->entries[i].value,
	     customFunctionsNamesD60[data->entries[i].index].name,
	     customFunctionsNamesD60[data->entries[i].index].options[data->entries[i].value]
	     );
    }
  }

}


void Ciff_Parser::Print_Bytes_Decimal(int length, void *data)
{
  int i=0;

  while (i<length) {
    int j;
    printf("%s    ", outputPrefix);
    for (j=0;j <outputBytesPerRow && i < length;j++) {
      unsigned char value = ((unsigned char*)data)[i];
      printf("%3d ", value);
      i++;
    }
    printf("\n");
  }
}

void Ciff_Parser::Print_Int4_Decimal(int length, void *data)
{
  int i=0;
  int int4Length = length / 4;
  int4 *int4Data = (int4 *)data;

  while (i<int4Length) {
    int j;
    printf("%s    ", outputPrefix);
    for (j=0;j <outputInt4PerRow && i < int4Length;j++) {
      int value = Convert_Number_4_bytes(int4Data[i], pPhoto->byteOrder);

      printf("%10d ", value);
      i++;
    }
    printf("\n");
  }
}

void Ciff_Parser::Print_Float4(int length, void *data)
{
  int i=0;
  int floatLength = length / 4;
  float4 *ptr = (float4 *)data;
  assert(sizeof(float) == 4);

  while (i<floatLength) {
    int j;
    printf("%s    ", outputPrefix);
    for (j=0;j <outputInt4PerRow && i < floatLength;j++) {
      //      float temp = Convert_Number_Float4(ptr[i], pPhoto->byteOrder);
      float temp = Convert_Number_Float4(ptr[i], pPhoto->byteOrder);

      printf("%15g ", temp);
      i++;
    }
    printf("\n");
  }
}


void Ciff_Parser::Print_Int2_Decimal(int length, void *data)
{
  int i=0;
  int int2Length = length / 2;
  int2 *int2Data = (int2 *)data;

  while (i<int2Length) {
    int j;
    printf("%s    ", outputPrefix);
    for (j=0;j <outputInt2PerRow && i < int2Length;j++) {
      int value = Convert_Number_2_bytes(int2Data[i], pPhoto->byteOrder);

      printf("%6d ", value);
      i++;
    }
    printf("\n");
  }
}


void Ciff_Parser::Process_Record_Type_Binary(unsigned short type,
					    int length,
					    void *data)
{
  printf("%s   Values as bytes\n", outputPrefix);
  this->Print_Bytes_Hex(length,data);
  printf("%s   Values as bytes decimal\n", outputPrefix);
  this->Print_Bytes_Decimal(length,data);
  printf("%s   Values as int2\n", outputPrefix);
  this->Print_Int2_Decimal(length, data);
  printf("%s   Values as int4\n", outputPrefix);
  this->Print_Int4_Decimal(length, data);
  printf("%s   Values as float\n", outputPrefix);
  this->Print_Float4(length, data);

}

void Ciff_Parser::Process_Record_Type_Photo_Info(unsigned short type,
						 ciff_photo_info *photoInfo)
{
  /* THis could be done better with an array... but that is for the future */


  assert(photoInfo->continuousDrive < sizeof(focusingModesNames)/sizeof(char*));
  assert(photoInfo->meteringMode < sizeof(meteringModesNames)/sizeof(char*));
  assert(photoInfo->exposureMode < sizeof(exposureModesNames)/sizeof(char*));
  assert(photoInfo->macroMode < sizeof(macroModesNames)/sizeof(char*));
  assert(photoInfo->flashMode < sizeof(  flashModesNames)/sizeof(char*));
  assert(photoInfo->imageSize < sizeof(  imageSizesNames)/sizeof(char*));
  assert(photoInfo->easyShootingMode < sizeof( easyShootingModesNames)/sizeof(char*));
  assert(photoInfo->captureMode < sizeof( captureModesNames)/sizeof(char*));

  printf("%s   Macro Mode:               %s\n"
	 "%s   Self Timer:               %4.2f seconds\n"
	 "%s   Quality:                  %d \n"
	 "%s   Flash Mode:               %s\n"
	 "%s   Continuos Drive:          %d \n"
	 "%s   Unknown1:                 %d\n"
	 "%s   Focusing Mode:            %s\n"
	 "%s   Unknown2:                 %d\n"
	 "%s   Capture mode:             %s\n"
	 "%s   Image Size:               %s\n"
	 "%s   Easy shooting mode:       %s\n"
	 "%s   Digital Zoom:             %d\n"
	 "%s   image contrast:           %d image saturation %d image sharpness %d\n"
	 "%s   image ISO:                %d\n"
	 "%s   Metering mode:            %s\n"
	 "%s   Focusing Type:            %d\n"
	 "%s   Autofocus point selected: 0x%x \n"
	 "%s   exposure mode:            %s\n"
	 "%s   unknown4                  0x%04x\n"
	 "%s   unknown5                  0x%04x\n"
	 "%s   unknown6                  0x%04x\n"
	 "%s   unknown7                  0x%04x\n"
	 "%s   flashActivity             0x%04x\n"
	 "%s   flashDetails              0x%04x\n"

	 ,
	 outputPrefix,
	 macroModesNames[photoInfo->macroMode],

	 outputPrefix,
	 photoInfo->selfTimer * 0.1,

	 outputPrefix,
	 photoInfo->quality,

	 outputPrefix,
	 flashModesNames[photoInfo->flashMode],

	 outputPrefix,
	 photoInfo->continuousDrive,

	 outputPrefix,
	 photoInfo->unknown1,

	 outputPrefix,
	 focusingModesNames[photoInfo->focusingMode],

	 outputPrefix,
	 photoInfo->unknown2,

	 outputPrefix,
	 captureModesNames[photoInfo->captureMode],

	 outputPrefix,
	 imageSizesNames[photoInfo->imageSize],

	 outputPrefix,
	 easyShootingModesNames[photoInfo->easyShootingMode],

	 outputPrefix,
	 photoInfo->digitalZoom,

	 outputPrefix,
	 photoInfo->imageContrast,
	 photoInfo->imageSaturation,
	 photoInfo->imageSharpness,

	 outputPrefix,
	 photoInfo->imageISO,

	 outputPrefix,
	 meteringModesNames[photoInfo->meteringMode],
	 outputPrefix,

	 photoInfo->focusingType,

	 outputPrefix,
	 photoInfo->autoFocusPointSelected,

	 outputPrefix,
	 exposureModesNames[photoInfo->exposureMode],

	 outputPrefix,
	 photoInfo->unknown4,
	 outputPrefix,

	 photoInfo->unknown5,

	 outputPrefix,
	 photoInfo->unknown6,
	 outputPrefix,

	 photoInfo->unknown7,
	 outputPrefix,
	 photoInfo->flashActivity,
	 outputPrefix,
	 photoInfo->flashDetails
	 );

  if (photoInfo->lensFocalLengthMax == photoInfo->lensFocalLengthMin) {
    printf("%s   lens:                     %d mm\n",
	   outputPrefix, photoInfo->lensFocalLengthMax * photoInfo->lensFocalLengthUnits);

  } else {
    printf("%s   lens:                     %d-%d mm\n",
	   outputPrefix,
	   photoInfo->lensFocalLengthMin * photoInfo->lensFocalLengthUnits,
	   photoInfo->lensFocalLengthMax * photoInfo->lensFocalLengthUnits
	   );

  }


}



void Ciff_Parser::Process_Record_Type_Camera_Serial_Number(unsigned short type,
					       ciff_camera_serial_number *serialNumberRecord)
{
  printf("%s   Camera Serial number [%d]\n", outputPrefix, serialNumberRecord->cameraNumber);
}

void Ciff_Parser::Process_Record_Type_Focal_Length(unsigned short type,
						   ciff_focal_length *focalLength)
{
  printf("%s   Focal Length [%d] Size of Sensor %dx%d (1/1000 in)\n",
	 outputPrefix, focalLength->focalLength,
	 focalLength->xSizeOfSensor,
	 focalLength->ySizeOfSensor
	 );
}

void Ciff_Parser::Process_Record_Type_Exposure_Info(unsigned short type,
						    ciff_translated_photo_exposure_information* exposure)
{
  /* THe exposure is dynamically allocated and it is freed after this
     function */

  /* Exposure time is in seconds, but we don't like to see it in seconds when it is
     less than 1 s */

  double printableTargetTV;
  double printableActualTV;

  /*
     TODO. Look into this, because the numbers are not "exactly" what photograhers
     are used to. Maybe we need a lookup table
  */

  if (exposure->targetTV < 1) {
    printableTargetTV = 1/exposure->targetTV;
  } else {
    printableTargetTV = exposure->targetTV;
  }

  if (exposure->actualTV < 1) {
    printableActualTV = 1/exposure->actualTV;
  } else {
    printableActualTV = exposure->actualTV;
  }

  printf("%s   Exposure iso:         %d\n"
	 "%s   Target AV:            %5.1f \n"
	 "%s   Target TV:                 %12.1f \n"
	 "%s   ExposureCompensation: %d\n"
	 "%s   White Balance:        %s\n"
	 "%s   Auto Exposure Comp:   %x\n"
	 "%s   Flash Exposure Comp:  %d \n"
	 "%s   Auto Bracketing Comp: %d \n"
	 "%s   Actual AV:            %5.1f\n"
	 "%s   Actual TV:            %12.1f \n",
	 outputPrefix,
	 exposure->iso,

	 outputPrefix,
	 exposure->targetAV,

	 outputPrefix,
	 printableTargetTV,

	 outputPrefix,
	 exposure->exposureCompensation,
	 /* 32 -> +1 */
	 outputPrefix,
	 whiteBalanceNames[exposure->whiteBalance],

	 outputPrefix,
	 exposure->autoExposureCompensation,
	 /* 0x3000 -> +1, 3002 -> 0 */

	 outputPrefix,
	 exposure->flashExposureCompensation,

	 /* Value of 20 for 2/3 */

	 outputPrefix,
	 exposure->autoExposureBracketingCompensation,

	 outputPrefix,
	 exposure->actualAV,

	 outputPrefix,
	 printableActualTV
	 );
}

void Ciff_Parser::Process_Record_Type_Photo_Serial_Number(unsigned short type,
							  ciff_translated_photo_serial_number *serialNumber)
{

  printf("%s    Photo serial number: %d\n",
	 outputPrefix, serialNumber->serialNumber);
}


void Ciff_Parser::Process_Record_Type_Image_Format(unsigned short type,
						   ciff_translated_image_format *imageFormat)
{
  /* Make sure we don't try to  beyond boundaries */
  assert(imageFormat->compressionType < sizeof(compressionTypesNames)/sizeof(char*));
  assert(imageFormat->fileFormat < sizeof(fileFormatsNames)/sizeof(char*) );

  printf("%s    File Format: %s Compression Type: %s Compression Ration %12.3f\n",
	 outputPrefix,
	 fileFormatsNames[imageFormat->fileFormat],
	 compressionTypesNames[imageFormat->compressionType],
	 imageFormat->targetCompressionRatio);
}

void Ciff_Parser::Process_Record_Type_Image_Details(unsigned short type,
						    ciff_image_details *imageDetails)
{
  /* ColorBW:

  I am too lazy to do this now.
  LSB: 0 grayscale, 1 color
  2nd byte: Does it need postprocessign pixel aspect ratio? 0 no, 1 yes

  */

  printf("%s    imageWidth: %d imageHeight: %d\n"
	 "%s     pixelAspectRatio: %f rotationAngle: %d\n "
	 "%s    component bitDepth: %d color bitDepth: %d color BW: %d\n",
	 outputPrefix,
	 imageDetails->imageWidth,
	 imageDetails->imageHeight,

	 outputPrefix,
	 imageDetails->pixelAspectRatio,
	 imageDetails->rotationAngle,

	 outputPrefix,
	 imageDetails->componentBitDepth,
	 imageDetails->colorBitDepth,
	 imageDetails->colorBW);
}


void Ciff_Parser::Process_Record_Type_Image_Details(unsigned short type,
					 ciff_target_image_type *targetImage)

{
  assert(targetImage->targetImageType < sizeof(targetImageTypesNames)/sizeof(char*));

  printf("%s    target image type: %s\n",
	 outputPrefix,
	 targetImageTypesNames[targetImage->targetImageType]);

}

void Ciff_Parser::Process_Record_Header(unsigned short type,
					int length,
					int recordOffset,
					int dataOffset)
{
  char *name;
  char marker;

  if ((type & 0xFF00) == 0x3000  ||
      (type & 0xFF00) == 0x2800 )
    marker = '+';
  else
    marker = '|';

  name = this->Find_Record_Name(type);
  printf( "%s%c %-20s 0x%04x len %8d recOff %8d dataOff %8d\n",
	  outputPrefix, marker,
	  name, type, length, recordOffset, dataOffset);

}


void Ciff_Parser::Process_Record(unsigned short type, int length,
				 char *dataInHeader,
				 int recordOffset,
				 int dataOffset)
{



   /* The data might NOT have been read

   */

   memset(outputPrefix, ' ', 256);
   assert(recursionLevel < 256/2);
   outputPrefix[(recursionLevel-1)*2] = 0;

   this->Process_Record_Header(type, length, recordOffset, dataOffset);


   if (length == 8) {
     Print_Bytes_Hex(length, dataInHeader);
   }
   switch (type) {
   case RECORD_CAMERA_SERIALNUMBER:
     {
       ciff_camera_serial_number *serialNumberRecord;

       serialNumberRecord = (ciff_camera_serial_number *)dataInHeader;

       Fix_Endianism_In_Record(dataInHeader, CIFF_CAMERA_SERIAL_NUMBER_RECORD_SPEC, length, pPhoto->byteOrder);
       this->Process_Record_Type_Camera_Serial_Number(type, serialNumberRecord);

     }
     break;
   case RECORD_FOCAL_LENGTH:
     {
       ciff_focal_length *focalLengthRecord;
       focalLengthRecord = (ciff_focal_length *)dataInHeader;
       Fix_Endianism_In_Record(dataInHeader, CIFF_FOCAL_LENGTH_RECORD_SPEC, length, pPhoto->byteOrder);
       this->Process_Record_Type_Focal_Length(type, focalLengthRecord);
     }
     break;
   case RECORD_CUSTOM_FUNCTIONS:
     {
       void *data=NULL;
       ciff_translated_custom_functions *customFNdata=NULL;
       data = this->Read_Data(length, dataOffset);

       assert(data != NULL);


       this->Translate_Custom_Functions(&customFNdata, data);

       assert(customFNdata != NULL);

       Process_Record_Custom_Functions(type, customFNdata);

       free(customFNdata);

       free(data);

     }
     break;
   case RECORD_EXPOSURE_INFO:
     {
       ciff_photo_exposure_information *exposure;
       ciff_translated_photo_exposure_information goodExposure;

       exposure = (ciff_photo_exposure_information *)this->Read_Data(length, dataOffset);

       Process_Record_Type_Binary(type, length, exposure);

       Fix_Endianism_In_Record(exposure, CIFF_PHOTO_EXPOSURE_INFORMATION_SPEC, length, pPhoto->byteOrder);

       this->Translate_Exposure_Record(&goodExposure, exposure);

       Process_Record_Type_Exposure_Info(type, &goodExposure);


       free(exposure);
     }
     break;
   case RECORD_IMAGE_DETAILS:
     {
       assert(length = sizeof(ciff_image_details));
       ciff_image_details *imageDetails;

       imageDetails = (ciff_image_details*)this->Read_Data(length, dataOffset);

       Fix_Endianism_In_Record(imageDetails, CIFF_IMAGE_DETAILS_RECORD_SPEC, length, pPhoto->byteOrder);

       Process_Record_Type_Image_Details(type, imageDetails);

       free(imageDetails);
     }
     break;
   case RECORD_TARGET_IMAGE_TYPE:
     {
       ciff_target_image_type targetImage;
       assert(sizeof(targetImage) == 8);
       memcpy(&targetImage, dataInHeader, 8);
       Fix_Endianism_In_Record((void*)&targetImage, CIFF_TARGET_IMAGE_RECORD_SPEC, length, pPhoto->byteOrder);

       Process_Record_Type_Image_Details(type, &targetImage);
     }
     break;

   case RECORD_IMAGE_DATE:
     {
       ciff_date *dateRecord;

       dateRecord = (ciff_date*) this->Read_Data(length, dataOffset);
       if (dateRecord != NULL) {

	 Fix_Endianism_In_Record((void*)dateRecord, CIFF_DATE_RECORD_SPEC, length, pPhoto->byteOrder);
	 this->Process_Record_Type_Date(type, dateRecord);
	 pPhoto->imageDate = dateRecord->dateTT;

	 free(dateRecord);
       }
     }
     break;
   case RECORD_PHOTO_SERIALNUMBER:
     {
       ciff_translated_photo_serial_number serialNumber;
       Translate_Photo_Serial_Number_Record(&serialNumber, dataInHeader);
       Process_Record_Type_Photo_Serial_Number(type, &serialNumber);

     }
     break;
   case RECORD_IMAGE_FORMAT:
     {
       ciff_translated_image_format imageFormat;
       Translate_Image_Format(&imageFormat, dataInHeader);

       Process_Record_Type_Image_Format(type, &imageFormat);
     }
     break;

   case RECORD_IMAGE_DESCRIPTION:
   case RECORD_UNKNOWN11:
   case RECORD_UNKNOWN12:
   case RECORD_UNKNOWN14:
   case RECORD_UNKNOWN15:
   case RECORD_WHITE_BALANCE:
     {
       unsigned char *data;

       data = (unsigned char*) this->Read_Data(length, dataOffset);

       if (data != NULL) {

	 Print_Int2_Decimal(length, data);
	 free(data);
       }

     }
     break;
   case RECORD_ROM_OP_MODE:
     {
       Process_Record_Type_String(type, dataInHeader);
     }
     break;
   case RECORD_CAMERA_NAME:
     {
       /* FIx this */
       char *str;
       str = (char*)this->Read_Data(length, dataOffset);
       strcpy(pPhoto->cameraName, str+6);
       if (str[5] == '\0') str[5] = ',';
       if (str != NULL) {
	 Process_Record_Type_String(type, str);
	 free(str);
       }
     }
     break;
   case RECORD_PHOTO_INFO:
     {

       ciff_photo_info *photoInfo;
       photoInfo = (ciff_photo_info*) this->Read_Data(length, dataOffset);

       Fix_Endianism_In_Record((void*)photoInfo, CIFF_PHOTO_INFO_SPEC, length, pPhoto->byteOrder);
       this->Process_Record_Type_Photo_Info(type, photoInfo);

       free(photoInfo);
     }
     break;

   }




   if ((type & 0xFF00) == 0x800 && type != RECORD_CAMERA_NAME) {
     char *str;
     str = (char*)this->Read_Data(length, dataOffset);
     if (str != NULL) {
       //       Process_Record_Type_Binary(type, length, str);
       Process_Record_Type_String(type, str);
       free(str);
     }
   }

   /*
     else if ((type & 0xFF00) == 0x1800) {

     void *data = this->Read_Data(length, dataOffset);
     if (data != NULL) {
       Process_Record_Type_Binary(type, length, data);
       free(data);
     }
   */
   /*
   if (dataInHeader)
     Process_Record_Type_Binary(type, length, dataInHeader);
   */

}


void Ciff_Parser::Set_Photo(craw_photo_type *pInitPhoto)
{
  this->pPhoto = pInitPhoto;
}

int Ciff_Parser::Ciff_Read_Records_In_Heap_Recursively(long dataOffset,
						       long lengthData)
{

/*

   Heaps are composed of a sequence of records.

   First we need to retrieve the offset of the data section of the heap.

   The data section contains:

   2 bytes: number of records in this heap

   For each record in the heap

   * Type of record

   * Data for the record:

     - If the record type is mask 0x4000, then it is 8 bytes long, and
     those bytes are stored right here.

     - Otherwise we need to read 4 bytes that contain its the file
       offset where the actual record is stored.

   * If the record is of certain type (?????) then its data
     is a heap and should be read recursively.

*/
   int count = 0;
   int i;
   int returnValue = 0;


   long recordsOffset;
   /*   unsigned long dummyLong; */


   /* Let us protect ourselves from unlimited recursion, 10 seems to be a good value */

   assert(recursionLevel++ < 10);

   /* A group of records starts in recordsOffset.
      The number of records is at the _END_ */


   /* First we retrieve the location of the data record for this heap */

   /* read the last 4 bytes from the data block */
   fseek(pPhoto->file, dataOffset+lengthData - 4, SEEK_SET);

   recordsOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder);


   /**** We now read the data, which is located after the headers */

   /* First 2 bytes are the number of records */

   fseek(pPhoto->file, recordsOffset + dataOffset, SEEK_SET);
   count = Read_Int_2(pPhoto->file, pPhoto->byteOrder);

   /*
   DEBUG_3( "%s> Number of Records [%d]\n", prefix, count);
   */
   /* let us make sure we don't have weird values */
   assert(count> 0 && count <100);

   /* Process each of the records */

   for (i=0;i<count;i++) {
      short type =-1;
      int length=-1;
      int offsetDataRecord = -1;
      long saveFileLocation = -1;
      char *data = NULL;
      char dummy8[8];

      int currentOffset = ftell(pPhoto->file);

      type = Read_Int_2(pPhoto->file, pPhoto->byteOrder);

      if (type & 0x4000) {
	 length = 8;
	 fread(dummy8, length, 1, pPhoto->file);
	 data = dummy8;
      } else {
	 length = Read_Int_4(pPhoto->file, pPhoto->byteOrder);
	 offsetDataRecord = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;
      }

      saveFileLocation = ftell(pPhoto->file);
      if (Act_On_Record(type)) {

	this->Process_Record(type, length, data, currentOffset, offsetDataRecord);

      }
      fseek(pPhoto->file, saveFileLocation, SEEK_SET);


      switch (type) {
      case RECORD_CAMERA_HEAP: /* Camera object */
      case RECORD_IMAGE_DETAILS_HEAP: /* Image spec */
      case 0x3003: /* Measured info */
      case 0x3005:
      case 0x300a:
      case 0x3004: /* Camera spec? */
      case RECORD_CAMERA_SETUP_HEAP:
	saveFileLocation = ftell(pPhoto->file);
	returnValue = this->Ciff_Read_Records_In_Heap_Recursively(offsetDataRecord, length);
	fseek(pPhoto->file, saveFileLocation, SEEK_SET);
	break;
      }
   }
   return returnValue;

#ifdef ALL_THIS_CODE_HAS_BEEN_REMOVED
      /*
      if (debug) {
	 fprintf(ERROR_OUTPUT, "%sREading record %5d %5x type %5d length %8d %8x\n",
		 prefix,
		 i, type, type, length, length);
      }
      */


      switch (type) {
      case RECORD_RAW:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;
	 DEBUG_6( "%s*Raw photo [%x] [%d] [%x] %d\n",prefix, recordOffset, recordOffset,
		 length, length);
	 break;
      case RECORD_JPEG:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) +  dataOffset;

	 pPhoto->jpegOffset = recordOffset;
	 pPhoto->jpegLength = length;

	 DEBUG_5("%s*Embedded jpeg image %d [%x] %d\n",prefix, recordOffset, recordOffset, length);
	 break;
      case RECORD_CAMERA_NAME:
	 DEBUG_4("%sCAMERA NAME [%x %d]\n",prefix, length, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;

	 Set_Camera_Name(pPhoto, length, recordOffset, dataOffset);

	 break;
      case RECORD_EXPOSURE_INFO: /* Exposure information */
	 DEBUG_5( "%sEXPOSURE INFORMATION [%x] %d %x\n",prefix, type, length, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;

	 {
	    photo_exposure_information_ciff photoExposureData;
	    /*

 2 iso
 3 exposure acuracy
 4 targetAV
 5 targetTV
 6 exposureCompensation: 32 for +1
 7 White balance
 8 unknown_1    3 --- ALl photos have a 3
 9 unknown_2    0 --all photos
10 unknown_3    0 --all photos
11 unknown_4    8 --all photos
12 unknown_5    0 --all photos
13 unknown_6    ---->   0 for flash off --> 0xffff for flash on
14 autoExpCompensaion 3002, 3000, 30004, 3004 ??? 3001 with AEB (-4/5, +4/5)
15 flashExposureCompensation   -> ffc0 -> -2     (-64)
				  ffcc -> -5/3   (-52)
				  ffd4 -> -4/3   (-44)
				  ffe0 -> -1     (-32)
				  ffec -> -2/3   (-20)
				  fff4 -> -1/3   (-12)
				  0    ->  0
				  32   ->  1
				  64   ->  2
16              0 --all photos
17 autoBraketingCompensation    (same as 15)
18              1 --all photos
19 minDistanceSubject               --autofocus related  0xffff (-1) Manual Focus
20 maxDistanceSubject               --autofocus related  0xffff (-1) and infiity, zero why?
21 actualAV
22 actualTV
23 exposure?
24 unknwon          0   all photos
25 unknwon          0   all photos
26 unknown          252 all photos
	    */


	    ciff_Read_Record_Data(pPhoto, &photoExposureData, recordOffset, length);


#define MOVE_INT2(a,b) pPhoto->a = Convert_Number_2_bytes(photoExposureData.b, pPhoto->byteOrder)
	    MOVE_INT2(imageISO, iso);
	    MOVE_INT2(imageAperture, targetAV);
	    MOVE_INT2(imageTimeExposure, targetTV);
	    MOVE_INT2(imageWhiteBalance, whiteBalance);
	    MOVE_INT2(flashExposureCompensation, flashExposureCompensation);
	    MOVE_INT2(autoExposureBracketingCompensation, autoExposureBracketingCompensation);
	    MOVE_INT2(imageActualAperture, actualAV);
	    MOVE_INT2(imageActualTimeExposure, actualTV);


#undef MOVE_INT2
	    switch (pPhoto->imageISO) {
	    case 160:
	       pPhoto->imageISO = 100;
	       break;
	    case 192:
	       pPhoto->imageISO = 200;
	       break;
	    case 224:
		  pPhoto->imageISO = 400;
	       break;
	    case 256:
		  pPhoto->imageISO = 800;
	       break;
	    default:
	       DEBUG_2( "*******Not dealing with this ISO yet %d\n", pPhoto->imageISO);
	    }

	    DEBUG_10("                              ISO %d Aperture %d Time Exposure %d FLash mode %d\n"
		    "                              White balance %d Flash Exp. Compensation %d Auto Exposure Bracketing %d\n"
		    "                              Acutal Aperture %d Actual TIme Exposure %d \n",
		    pPhoto->imageISO,
		    pPhoto->imageAperture,
		    pPhoto->imageTimeExposure,
		    pPhoto->flashMode,

		    pPhoto->imageWhiteBalance,
		    pPhoto->flashExposureCompensation,
		    pPhoto->autoExposureBracketingCompensation,

		    pPhoto->imageActualAperture,
		    pPhoto->imageActualTimeExposure
		    );

	    Print_Int2(&photoExposureData, length, pPhoto->byteOrder);

	 }

	 break;
      case RECORD_CUSTOM_FUNCTIONS:
	 DEBUG_5( "%sCFs [%x] %d %x\n",prefix, type, length, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;

	 if (strcmp(pPhoto->cameraName, "Canon EOS D60")==0) {
	    Read_CF(pPhoto, length, recordOffset, dataOffset);
	 }

	 break;
      case RECORD_CAMERA_FIRMWARE:

	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;

	 assert(length < 256);
	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);

	 DEBUG_6( "%sFIRMWARE ID [%x] %d %x FIrmware [%s]\n",prefix, type, length, length, buffer);

	 break;
      case RECORD_EXIF_EXPOSURE:
	 DEBUG_4( "%sEXIF EXPOSURE DATA [%x] Length [%d]\n",prefix, type, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;
	 {
	    photo_exposure_EXIF_ciff photoExposureData;

	    ciff_Read_Record_Data(pPhoto, &photoExposureData, recordOffset, length);

#define MOVE_INT2(a,b) pPhoto->a = Convert_Number_2_bytes(photoExposureData.b, pPhoto->byteOrder)
	    pPhoto->macroMode= Convert_Number_2_bytes(photoExposureData.macroMode, pPhoto->byteOrder);
	    pPhoto->selfTimer= Convert_Number_2_bytes(photoExposureData.selfTimer, pPhoto->byteOrder);
	    pPhoto->imageQuality= Convert_Number_2_bytes(photoExposureData.quality, pPhoto->byteOrder);
	    pPhoto->flashMode= Convert_Number_2_bytes(photoExposureData.flashMode, pPhoto->byteOrder);
	    pPhoto->continuousDrive = Convert_Number_2_bytes(photoExposureData.continuousDrive, pPhoto->byteOrder);

	    MOVE_INT2(focusingMode, focusingMode);

	    pPhoto->imageSizeRelative = Convert_Number_2_bytes(photoExposureData.imageSize, pPhoto->byteOrder);
	    pPhoto->easyProgramMode = Convert_Number_2_bytes(photoExposureData.easyShootingMode, pPhoto->byteOrder);
	    pPhoto->digitalZoom  = Convert_Number_2_bytes(photoExposureData.digitalZoom, pPhoto->byteOrder);
	    MOVE_INT2(imageContrast, imageContrast);
	    MOVE_INT2(imageSaturation, imageSaturation);
	    MOVE_INT2(imageSharpness, imageSharpness);
	    /*	    THis value is not reliable
	    MOVE_INT2(imageISO, imageISO);
	    */
	    MOVE_INT2(meteringMode, meteringMode);
	    MOVE_INT2(focusingType, focusingMode);
	    MOVE_INT2(autoFocusPointSelected, autoFocusPointSelected);
	    MOVE_INT2(exposureMode, exposureMode);
	    MOVE_INT2(lensFocalLengthMin, lensFocalLengthMin);
	    MOVE_INT2(lensFocalLengthMax, lensFocalLengthMax);
	    MOVE_INT2(lensFocalLengthUnits, lensFocalLengthUnits);

	    MOVE_INT2(flashInfo1, flashInfo1);
	    MOVE_INT2(flashInfo2, flashInfo2);
	    MOVE_INT2(flashActivity, flashActivity);
	    MOVE_INT2(flashDetails, flashDetails);


#undef MOVE_INT2

	    DEBUG_5("                             macro mode %d selftimelength %d quality %d FLash mode %d\n",
		    pPhoto->macroMode,
		    pPhoto->selfTimer,
		    pPhoto->imageQuality,
		    pPhoto->flashMode
		    );
	    DEBUG_5("                             continuos drive %d focus mode %d image size %d Program mode %d\n",
		    pPhoto->continuousDrive,
		    pPhoto->focusingMode,
		    pPhoto->imageSizeRelative,
		    pPhoto->easyProgramMode
		    );
	    DEBUG_5("                             digital zoom %d image conrast %d image saturation %d image sharpness %d\n",
		    pPhoto->digitalZoom,
		    pPhoto->imageContrast,
		    pPhoto->imageSaturation,
		    pPhoto->imageSharpness
		    );
	    DEBUG_5("                             ISO %d Metering mode %d focusing type %d Auto focus point selected %x\n",
		    pPhoto->imageISO,
		    pPhoto->meteringMode,
		    pPhoto->focusingType,
		    pPhoto->autoFocusPointSelected
		    );
	    DEBUG_5("                             Exposure Mode %d Focal Length Min %d Max %d Units per mm %d\n",
		    pPhoto->exposureMode,
		    pPhoto->lensFocalLengthMin,
		    pPhoto->lensFocalLengthMax,
		    pPhoto->lensFocalLengthUnits
		    );
	    DEBUG_8("                             flash info1 0x%x %d flash info 2 0x%x %d flashActivity 0x%x %d flashDetails %x\n",
		    pPhoto->flashInfo1, pPhoto->flashInfo1,
		    pPhoto->flashInfo2, pPhoto->flashInfo2,
		    pPhoto->flashActivity, pPhoto->flashActivity,
		    pPhoto->flashDetails
		    );
	    /*
	    DEBUG_6("                             UNKNOWNS u1 %u u2 %u u3 %u u4 %u u5 %u\n",
		    photoExposureData.unknown1,
		    photoExposureData.unknown2,
		    photoExposureData.unknown3,
		    photoExposureData.unknown4,
		    photoExposureData.unknown5
		    );

	    */

	    Print_Bytes((char*)&photoExposureData, length);

	    Print_Int2(&photoExposureData, length, pPhoto->byteOrder);
	 }
	 break;
      case 0x0001: /* D30 */
      case 0x0032: /* D30  Canon Color Info1*/
      case 0x0036: /* this record is identical in every photo I take*/
	 debug = 1;
	 DEBUG_4( "%s Color info? [%x] Length [%d]\n",prefix, type, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;
	 {
	    char buffer[8192];
	    assert(length <= 8192);
	    ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);

	    if (debug)
	       Print_Bytes(buffer, length);
	    /*
	    Print_Int2(buffer, length, pPhoto->byteOrder);
	    */

	 }
	 debug = 0;
	 break;
      case 0x10c2:
	 /* NO idea what this record is, but it is the same in all the photos */
	 DEBUG_1("Printing\n");
	 DEBUG_4( "%sUnknown [%x] Length [%d]\n",prefix, type, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;
	 if (debug)
	    Print_Bytes(buffer, length);
	 break;
      case 0x102c: /* D30 only CanonColorInfo2 */

      case 0x107f: /* G2 only */
      case 0x10a8:
      case 0x10a9: /* Whitebalance Table */
      case 0x10aa:
      case 0x10c0:
      case 0x10c1:
      case 0x1093: /* d300 only */
      case 0x1038: /* d300 only   CanonPictureInfo */
      case 0x10ad: /* d300 only */
      case 0x10ae: /* d300 only */
      case 0x10af: /* d300 only */
      case 0x10b4: /* d300 only  Color Space --- 1 sRGB, 2 Adobe RGV, 0xfff uncalibrated */
      case 0x10b5: /* d300 only */
      case 0x1039: /* d300 only */
      case 0x183b: /* d300 only */
      case 0x1833:
      case 0x1030: /* D30 only */
      case 0x1835: /* Decoder table */
	 DEBUG_4( "%s*Record LONGS [%x] Length [%d]\n",prefix, type, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;
	 {
	    char buffer[4086];
	    assert(length < 4086);
	    ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);

	    if (debug) Print_Bytes(buffer, length);

	 }
	 break;

      case 0x2008: /* G2 only */

	 DEBUG_4( "%s*Record LONGS [%x] Length [%d]\n",prefix, type, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;


	 /*
	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);
	 Print_Bytes(buffer, length);
	 */
	 /*
	 for (j=0;j<length/4;j++){
	    dummyLong = Read_Int_4(pPhoto->file, pPhoto->byteOrder);
	    DEBUG_4( "%s    %2d Long %8lx %10lu\n", prefix, j, dummyLong, dummyLong);
	 }
	 */
	 break;
      case RECORD_SENSOR_DATA:
	 DEBUG_5( "%sSENSOR [%x] %d %x\n",prefix, type, length, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;

	 assert(length < 256);
	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);
	 /*
	 Print_Int2(buffer, length, pPhoto->byteOrder);
	 */

	 break;

      case RECORD_IMAGE_DATE:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;

	 {
	    date_ciff dateRecord;


	    assert(length == sizeof(dateRecord));

	    ciff_Read_Record_Data(pPhoto,&dateRecord, recordOffset, length);

	    dateTT = Convert_Number_4_bytes(dateRecord.dateTT, pPhoto->byteOrder);

	    DEBUG_7( "%sDATE [%x] %d %x TT [%lu] %s",prefix, type, length, length, dateTT, ctime(&dateTT));
	    pPhoto->imageDate = dateTT;
	 }

	 break;
      case RECORD_IMAGE_DESCRIPTION:

	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;


	 assert(length <= 256);
	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);

	 DEBUG_5( "%sIMAGE DESCRIPTION [%x] Length %d Value [%s]\n",prefix, type, length, buffer);

	 /*
	 for (j=0;j<length/4;j++){
	    dummyLong = Read_Int_4(pPhoto->file, pPhoto->byteOrder);
	    DEBUG_5( "%s    %2d Long %8lx %10lu\n", prefix, j, dummyLong, dummyLong);
	 }
	 */
	 break;

      case RECORD_IMAGE_DETAILS:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;

	 {
	    image_description_ciff imageDescription;

	    assert(length == sizeof(image_description_ciff));
	    ciff_Read_Record_Data(pPhoto, &imageDescription, recordOffset, length);

	    pPhoto->imageXSize = Convert_Number_4_bytes(imageDescription.xSize, pPhoto->byteOrder);
	    pPhoto->imageYSize = Convert_Number_4_bytes(imageDescription.ySize, pPhoto->byteOrder);

	    DEBUG_7( "%sIMAGE DESCRIPTION [%x] %d %x X [%d] Y [%d] \n",
		    prefix, type, length, length, pPhoto->imageXSize, pPhoto->imageYSize);

	    /*
	      Print_Bytes(imageDescription.unknown1, sizeof(imageDescription.unknown1));
	    */
	 }


	 break;


      case RECORD_IMAGE_NAME: /* IMAGE_NAME  */
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;

	 assert(length < 256);

	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);

	 DEBUG_6( "%sIMAGE NAME [%x] %d %d [%s]\n",prefix, type, length, length, buffer);

	 break;



      case RECORD_FILENAME:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) +  dataOffset;
	 assert(length < 256);

	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);
	 strcpy(pPhoto->imageName, buffer);
	 DEBUG_6( "%sFilename [%x] %d %x [%s]\n",prefix, type, length, length, pPhoto->imageName);

	 break;

      case RECORD_OWNER_NAME:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) +  dataOffset;
	 assert(length < 256);
	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);

	 DEBUG_6( "%sOWNER NAME [%x] %d %x [%s]\n",prefix, type, length, length, buffer);

	 break;


      case RECORD_THM_FILENAME:

	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) +  dataOffset;
	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);
	 assert(length < 256);
	 DEBUG_6( "%sTHM filename [%x] %d %u [%s]\n",prefix, type, length, length, buffer);



	 break;
      case RECORD_UNKNOWN1:

	 DEBUG_5( "%s*XXXXXXXXXXX[%x] %d %d\n",prefix, type, length, length);
	 fread(dummy, 8, 1, pPhoto->file);
	 if (debug) Print_Bytes(dummy, 8);
	 break;

      case RECORD_FOCAL_LENGTH: /* Focal length */
	 {
	    focal_length_ciff focalLengthRaw;
	    fread(&focalLengthRaw, 8, 1, pPhoto->file);
	    pPhoto->imageFocalLength = Convert_Number_2_bytes(focalLengthRaw.focalLength, pPhoto->byteOrder);
	    DEBUG_5( "%sFOCAL Length[%x] %d FOCAL LENGTH %dmm\n",prefix, type, length, pPhoto->imageFocalLength);
	    if (debug) Print_Int2((char*)&focalLengthRaw, sizeof(focalLengthRaw),pPhoto->byteOrder);
	 }
	 break;

      case RECORD_PHOTO_SERIALNUMBER:
	 {
	    long serialNumber;
	    unsigned long unknown;

	    serialNumber = Read_Int_4(pPhoto->file, pPhoto->byteOrder);
	    unknown = Read_Int_4(pPhoto->file, pPhoto->byteOrder);

	    DEBUG_6( "%sSERIAL NUMBER[%x] %d Photo Serial Number %ld Unknown %lu\n",
		    prefix, type, length, serialNumber, unknown);

	    pPhoto->imageSerialNumber = serialNumber;

	    /*
	    Print_Bytes((char*)&unknown, 4);
	    Print_Int2(&unknown, 4, pPhoto->byteOrder);
	    */
	 }


	 break;

      case RECORD_CAMERA_NUMBER:
	 {
	    unsigned long cameraNumber;
	    unsigned long unknown;

	    cameraNumber = Read_Int_4(pPhoto->file, pPhoto->byteOrder);
	    unknown = Read_Int_4(pPhoto->file, pPhoto->byteOrder);

	    DEBUG_6( "%sCAMERA NUMBER[%x] %d Number %lu Unknown %lu\n",
		    prefix, type, length, cameraNumber, unknown);

	    pPhoto->cameraNumber = cameraNumber;

	 }


	 break;
      case 0x5814:
      case 0x5834: /* Model number? */
      case 0x501c:
      case RECORD_UNKNOWN1:
      case RECORD_UNKNOWN2:
      case RECORD_UNKNOWN3:
      case 0x480d:

	 debug=1;

	 DEBUG_5( "%s*Record UNKNOWN[%x] %d %d\n",prefix, type, length, length);
	 fread(dummy, 8, 1, pPhoto->file);

	 if (debug) {
	    Print_Bytes(dummy, 8);
	    Print_Int2(dummy, 8, pPhoto->byteOrder);
	    Print_Int4(dummy, 8, pPhoto->byteOrder);
	 }
	 debug = 0;

	 break;

	 /* These are the recursive records: */
      case 0x2807: /* Camera object */
      case 0x2804: /* Image spec */
      case 0x3003: /* Measured info */
      case 0x3005:
      case 0x300a:
      case 0x3004: /* Camera spec? */
      case 0x300b:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset ;

	 /*
	   DEBUG_6( "%sRecursive record %d %x---------------------------------\n",prefix, recordOffset, recordOffset);
	 */

	 switch (type) {
	 case 0x2807: /* Camera object */
	    DEBUG_2( "%sCAMERA OBJECT\n", prefix);
	    break;
	 case 0x2804:
	    DEBUG_2( "%sIMAGE SPEC\n", prefix);
	    break;
	 case 0x3003: /* Measured info */
	    DEBUG_2( "%sMEASURED INFO\n", prefix);
	    break;
	 case 0x3004:
	    DEBUG_2( "%sCAMERA SPEC\n", prefix);
	    break;
	 case 0x3005:
	    break;
	 case 0x300a:
	    DEBUG_2( "%sIMAGE PROPERTIES\n", prefix);
	    break;
	 case 0x300b:
	    DEBUG_2( "%sRAW IMAGE PROPERTIES\n", prefix);
	    break;
	 }
	 saveFileLocation = ftell(pPhoto->file);
	 self->Read_Records_in_Heap__Recursively(pPhoto, recordOffset, length, level+1);
	 fseek(pPhoto->file, saveFileLocation, SEEK_SET);
	 break;
      default:
	 DEBUG_6( "REading record %5d Type [%5x] type [%5d] length %8d %8x\n",
		 i, type, type, length, length);

	 My_Die("Unsuported record\n");
      }





   }
#endif
  recursionLevel--;

}


int Ciff_Parser::Parse(void)
{

  return this->Ciff_Read_Records_In_Heap_Recursively(pPhoto->headerLength,
						     pPhoto->fileLength - pPhoto->headerLength);


}
