/*
   $Id: libciff.c,v 1.11 2004/08/03 03:24:17 dmg Exp $

   Library to read the records from a RAW file from Canon

   Based upon parse.c from David Coffin

   Credits:

     David Coffin:  Version 0.1 of this program is based on his parse.c
     David Burren:  Embedded exif record 
     Peter Galbavy: His documentation of the format helped a lot


 * Copyright (C) 2004 Daniel M. German
 *
 * Author: Daniel M. German dmg _at_ turingmachine _dot_ org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "libraw.h"
#include "libcraw.h"
#include "libciff.h"
#include "crawdebug.h"

/*  Constants related to the hader */


/* Masks */
#define CIFF_TYPE_MASK             0xc000  /* 1100 0000 0000 0000 */
#define CIFF_TYPE_IN_HEAP_SPACE    0x0000
#define CIFF_TYPE_IN_RECORD_ENTRY  0x1000


#define RECORD_RAW               0x2005
#define RECORD_JPEG              0x2007  

#define CIFF_MASK_STRING         0x0800   /* 0000 1000 */
#define CIFF_MASK_TYPE_RECORD    0x1000   /* 0001 0000 */
#define CIFF_TYPE_8_BYTES        0x5000   /* 1001 0000 */

#define RECORD_SENSOR_DATA       (CIFF_MASK_TYPE_RECORD | 0x0031)
#define RECORD_EXPOSURE_INFO     (CIFF_MASK_TYPE_RECORD | 0x002a)
#define RECORD_CUSTOM_FUNCTIONS  (CIFF_MASK_TYPE_RECORD | 0x0033)  /* Bunch of byte values */

#define RECORD_IMAGE_DETAILS     (CIFF_MASK_TYPE_RECORD | 0x0810)  /* INcludes rotarion information */
#define RECORD_IMAGE_DATE        (CIFF_MASK_TYPE_RECORD | 0x080e)

/* ---------------------------------------------------------------------- */

#define RECORD_FILENAME          (CIFF_MASK_STRING | 0x0016)
#define RECORD_THM_FILENAME      (CIFF_MASK_STRING | 0x0017)
#define RECORD_OWNER_NAME        (CIFF_MASK_STRING | 0x0010)

/* This appers to be a user filled field with the description of the photo*/
#define RECORD_IMAGE_DESCRIPTION (CIFF_MASK_STRING | 0x0005) 
#define RECORD_IMAGE_NAME        (CIFF_MASK_STRING | 0x0015) /* */

#define RECORD_CAMERA_NAME       (CIFF_MASK_STRING | 0x000a)
#define RECORD_CAMERA_FIRMWARE   (CIFF_MASK_STRING | 0x000b)


#define RECORD_UNKNOWN1           (CIFF_TYPE_8_BYTES | 0x000a) /* 8 Bytes, Image description? */
#define RECORD_FOCAL_LENGTH       (CIFF_TYPE_8_BYTES | 0x0029)

#define RECORD_UNKNOWN3           (CIFF_TYPE_8_BYTES | 0x0003)
#define RECORD_UNKNOWN2           (CIFF_TYPE_8_BYTES | 0x0004) /* 8 Bytes */
#define RECORD_PHOTO_SERIALNUMBER (CIFF_TYPE_8_BYTES | 0x0017) /* 8 Bytes */

#define RECORD_CAMERA_NUMBER      (CIFF_TYPE_8_BYTES | 0x000b)

#define RECORD_EXIF_EXPOSURE      (CIFF_TYPE_8_BYTES | 0x002d)

/* These are native data structures as they appear in the file */

typedef struct {
   int4           xSize;
   int4           ySize;
   char           unknown1[20];
} image_description_ciff;

typedef struct {
   char           unknown1[2]; 
   int2           focalLength;
   char           unknown2[4];
} focal_length_ciff;

typedef struct {
   int4           dateTT; 
   char           unknown[8];
} date_ciff;

typedef struct {
   int4           cameraNumber; 
   char           unknown[4];
} camera_number_ciff;

typedef struct {
   unsigned short length;
   int2 macroMode;
   int2 selfTimer; /* Tenths of seconds */
   int2 quality;
   int2 flashMode;
   int2 continuousDrive;
   int2 unknown1;
   int2 focusingMode;
   int2 unknown2;
   int2 captureMode; /* -1 NA, 0 unknown, 1 still/jpeg, 2 CCDRAW, 3movie*/
   int2 imageSize;
   int2 easyShootingMode;
   int2 digitalZoom;
   int2 imageContrast;
   int2 imageSaturation;
   int2 imageSharpness;
   int2 imageISO;
   int2 meteringMode; 
   int2 focusingType;
   int2 autoFocusPointSelected;
   int2 exposureMode;
   int2 unknown4;
   int2 unknown5;
   int2 lensFocalLengthMax;
   int2 lensFocalLengthMin;
   int2 lensFocalLengthUnits;
   int2 flashInfo1;  /* It is still unknown, 128 in all pictures*/
   int2 flashInfo2;  /* SOme pictures it is 304, some 285 */
   int2 flashActivity; /* 133 for external flash, 255 for internal */
   int2 flashDetails;  /* 4008 for external flash, 2008 for internal flash */
                       /*Bits 15..0: (from David Burren)
			 14: external E-TTL
			 13: internal flash
			 11: FP sync used
			 7: 2nd("rear")-curtain sync used
			 4: FP sync enabled
			 other bits unknown
		       */
   unsigned short u5;
   unsigned short u6;
   unsigned short u7;
   unsigned short u8;
   unsigned short u9;
   unsigned short u10;
   unsigned short u11;
   unsigned short u12;
   unsigned short u13;
   unsigned short u14;
   unsigned short u15;
   unsigned short u16;
   unsigned short u17;
   unsigned short u18;
   unsigned short u19;
   unsigned short u20;
   unsigned short u21;
   unsigned short u22;
   unsigned short u23;
   unsigned short u24;
   unsigned short u25;
   unsigned short u26;
   unsigned short u27;
   unsigned short unknown[100];
} photo_exposure_EXIF_ciff;

typedef struct {
   int4 length;  
   int2 iso;
   int2 unknown1; /* Looks like focal length FL-1 in 17-40, but in
		     28-105 it was 108 for 48, 52 for 90 */
   int2 avTarget;
   int2 tvTarget; /* 5 */
   int2 exposureCompensation;
   int2 whiteBalance;
   
   int2 unknown2[5];
   
   int2 unknown3; /* 0 for flash off --> 0xffff for flash on*/
   int2 autoExposureCompensation;
   int2 flashExposureCompensation; /* 15 */

   int2 unknown4;

   int2 autoExposureBracketingCompensation;

   int2 unknown5[3];

   int2 actualAV;
   int2 actualTV;

   int2 unknown6[4];

} photo_exposure_information_ciff;


/* declare a generic type for the callback... now that I think about it, why don't we do it object oriented, doh */

typedef int(*callback_type)(void);

typedef struct {
   int recordId;            /* Number of record to "handle" */
   //   int (*callback)(craw_photo_type*);   /* callback for this record     */
   callback_type callback;   /* callback for this record     */
} ciff_callback;



int Debug_Record_Raw_Callback(craw_photo_type *pPhoto)
{
   
}

ciff_callback debugCallbacks[1] = { 
   {RECORD_RAW, (callback_type)Debug_Record_Raw_Callback},
};





static
int ciff_Read_Record_Data(craw_photo_type *pPhoto, void *buffer, int offset, int length)
{
   int saveFileLocation = ftell(pPhoto->file);
   int read;

   /*   fprintf(ERROR_OUTPUT, "Reading data Offset %d %x  Length %d, %x\n", offset, offset, length, length); */
   fseek(pPhoto->file, offset, SEEK_SET);
   read = fread(buffer, length, 1, pPhoto->file);

   fseek(pPhoto->file, saveFileLocation, SEEK_SET);

   return read;
}

static
void Set_Camera_Name(craw_photo_type *pPhoto, int length , int recordOffset, int dataOffset)
{
   unsigned char buffer[256];
   assert(length < 256);
   int leftBytes;
   unsigned char *pTemp;

   assert(length < 256);
   ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);
   
   strcpy(pPhoto->manufacturerName, buffer);
   
   DEBUG_2( "********************%s\n", pPhoto->manufacturerName);
   
   pTemp = buffer+ strlen(buffer)+1;

   strcpy(pPhoto->cameraName, pTemp);

   DEBUG_2("********************%s\n", pPhoto->cameraName);
	 
   pTemp = pTemp+ strlen(pTemp)+1;
	 
   leftBytes = length - 
      (strlen(pPhoto->cameraName) + strlen(pPhoto->manufacturerName) + 2);

   DEBUG_2( "****** %d\n", leftBytes); 
       
   /*
   for (j = 0;j<leftBytes;j++) {
      DEBUG_("                 %d %c %d %x\n", j, pTemp[j],pTemp[j],pTemp[j]);
   }
   */
}


static
void Read_CF(craw_photo_type *pPhoto, int length , int recordOffset, int dataOffset)
{
   unsigned char buffer[256];
   int j;
   int cfLen;
   int saveFileLocation = ftell(pPhoto->file);
   int cfIndex;
   int cfValue;

   assert(length < 256);
	 

   /*   DEBUG_( "Reading data Offset %d %x  Length %d, %x\n", offset, offset, length, length); */
   fseek(pPhoto->file, recordOffset, SEEK_SET);

   cfLen = Read_Int_4(pPhoto->file, pPhoto->byteOrder);

   DEBUG_2("                    Length of CF %d\n", cfLen);


   ciff_Read_Record_Data(pPhoto, buffer, recordOffset+4, cfLen);


   for (j = 0;j<cfLen;j+=2) {
      cfValue = buffer[j];
      cfIndex = buffer[j+1];
      DEBUG_5( "                         [%2d] Custom Function     %d Value %d %d\n", j/2, cfIndex, cfValue, cfIndex);

   }

   fseek(pPhoto->file, saveFileLocation, SEEK_SET);
}


/*
 *
 */
static
void Read_Records_Recursively(craw_photo_type *pPhoto, long dataOffset, 
				       long lengthData, 
				       int level) 
{
   int count = 0;
   int i;
   char prefix[256];
   char buffer[256];
   
   long saveFileLocation;
   long recordsOffset;
   /*   unsigned long dummyLong; */
   unsigned char dummy[8];
   
   /* A group of records starts in recordsOffset.
      The number of records is at the _END_ */

   if (debug) {
      memset(prefix, ' ', 256);
      prefix[level*6] = 0;

      /*
      fprintf(ERROR_OUTPUT, "%s level %d dataOffset %ld %lx lengthData %ld %lx\n",
	      prefix, level,
	      dataOffset, dataOffset, 
	      lengthData, lengthData);
      */
   }


   /*
   DEBUG_4( "%sSeeking %ld\n", prefix, dataOffset + lengthData - 4);
   */
   
   /* read the last 4 bytes from the data block */
   fseek(pPhoto->file, dataOffset+lengthData - 4, SEEK_SET);

   recordsOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder);

   /*
   DEBUG_5( "%sOffset of record header %ld %lx\n", prefix, recordsOffset,recordsOffset);
   */

   /* Let us protect ourselves from unlimited recursion */
   
   assert(level < 10);

   /*   photo_record *pRecords; */
   
   /* First 2 bytes are the number of records */


   fseek(pPhoto->file, recordsOffset + dataOffset, SEEK_SET);

   /*
   if (debug) DEBUG_( "%sSeek at %ld %lx\n", prefix, recordsOffset+dataOffset,recordsOffset+dataOffset);
   */

   count = Read_Int_2(pPhoto->file, pPhoto->byteOrder);

   DEBUG_4( "%s>Records [%d] %x\n", prefix, count, count);
   /* let us make sure we don't have weird values */
   assert(count> 0 && count <100);

   /* Allocate the records */

   for (i=0;i<count;i++) {
      int type =-1;
      int length=-1;
      int recordOffset=-1;
      
      type = Read_Int_2(pPhoto->file, pPhoto->byteOrder);

      if (type & 0x4000) {
	 /*
	 DEBUG_2( "%s-Record with known length\n", prefix);
	 */
	 length = 8;
      } else {
	 length = Read_Int_4(pPhoto->file, pPhoto->byteOrder);
      }



      /*
      if (debug) {
	 fprintf(ERROR_OUTPUT, "%sREading record %5d %5x type %5d length %8d %8x\n",
		 prefix, 
		 i, type, type, length, length);
      }
      */
      

      switch (type) {
      case RECORD_RAW:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;
	 DEBUG_6( "%s*Raw photo [%x] [%d] [%x] %d\n",prefix, recordOffset, recordOffset,
		 length, length);
	 break;
      case RECORD_JPEG:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) +  dataOffset;

	 pPhoto->jpegOffset = recordOffset;
	 pPhoto->jpegLength = length;

	 DEBUG_5("%s*Embedded jpeg image %d [%x] %d\n",prefix, recordOffset, recordOffset, length);
	 break;
      case RECORD_CAMERA_NAME:
	 DEBUG_4("%sCAMERA NAME [%x %d]\n",prefix, length, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;
	 
	 Set_Camera_Name(pPhoto, length, recordOffset, dataOffset);

	 break;
      case RECORD_EXPOSURE_INFO: /* Exposure information */
	 DEBUG_5( "%sEXPOSURE INFORMATION [%x] %d %x\n",prefix, type, length, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;

	 {
	    photo_exposure_information_ciff photoExposureData;
	    /*

 2 iso
 3 exposure acuracy
 4 avTarget
 5 tvTarget
 6 exposureCompensation: 32 for +1
 7 White balance
 8 unknown_1    3 --- ALl photos have a 3
 9 unknown_2    0 --all photos
10 unknown_3    0 --all photos
11 unknown_4    8 --all photos
12 unknown_5    0 --all photos
13 unknown_6    ---->   0 for flash off --> 0xffff for flash on
14 autoExpCompensaion 3002, 3000, 30004, 3004 ??? 3001 with AEB (-4/5, +4/5)
15 flashExposureCompensation   -> ffc0 -> -2     (-64)
                                  ffcc -> -5/3   (-52)
                                  ffd4 -> -4/3   (-44)
                                  ffe0 -> -1     (-32)
                                  ffec -> -2/3   (-20)
                                  fff4 -> -1/3   (-12)
                                  0    ->  0     
                                  32   ->  1
                                  64   ->  2
16              0 --all photos
17 autoBraketingCompensation    (same as 15)
18              1 --all photos
19 minDistanceSubject               --autofocus related  0xffff (-1) Manual Focus
20 maxDistanceSubject               --autofocus related  0xffff (-1) and infiity, zero why?
21 actualAV
22 actualTV
23 exposure?
24 unknwon          0   all photos
25 unknwon          0   all photos
26 unknown          252 all photos
	    */


	    ciff_Read_Record_Data(pPhoto, &photoExposureData, recordOffset, length); 


#define MOVE_INT2(a,b) pPhoto->a = Convert_Number_2_bytes(photoExposureData.b, pPhoto->byteOrder)
	    MOVE_INT2(imageISO, iso);
	    MOVE_INT2(imageAperture, avTarget);
	    MOVE_INT2(imageTimeExposure, tvTarget);
	    MOVE_INT2(imageWhiteBalance, whiteBalance);
	    MOVE_INT2(flashExposureCompensation, flashExposureCompensation);
	    MOVE_INT2(autoExposureBracketingCompensation, autoExposureBracketingCompensation);
	    MOVE_INT2(imageActualAperture, actualAV);
	    MOVE_INT2(imageActualTimeExposure, actualTV);	    


#undef MOVE_INT2
	    switch (pPhoto->imageISO) {
	    case 160:
	       pPhoto->imageISO = 100;
	       break;
	    case 192:
	       pPhoto->imageISO = 200;
	       break;
	    case 224:
		  pPhoto->imageISO = 400;
	       break;
	    case 256:
		  pPhoto->imageISO = 800;
	       break;
	    default:
	       DEBUG_2( "*******Not dealing with this ISO yet %d\n", pPhoto->imageISO);
	    }

	    DEBUG_10("                              ISO %d Aperture %d Time Exposure %d FLash mode %d\n"
		    "                              White balance %d Flash Exp. Compensation %d Auto Exposure Bracketing %d\n"
		    "                              Acutal Aperture %d Actual TIme Exposure %d \n",
		    pPhoto->imageISO,
		    pPhoto->imageAperture,
		    pPhoto->imageTimeExposure,
		    pPhoto->flashMode,

		    pPhoto->imageWhiteBalance,
		    pPhoto->flashExposureCompensation,
		    pPhoto->autoExposureBracketingCompensation,

		    pPhoto->imageActualAperture,
		    pPhoto->imageActualTimeExposure
		    );

	    if (debug)
	       Print_Int2(&photoExposureData, length, pPhoto->byteOrder);

	 }

	 break;
      case RECORD_CUSTOM_FUNCTIONS:
	 DEBUG_5( "%sCFs [%x] %d %x\n",prefix, type, length, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;

	 if (strcmp(pPhoto->cameraName, "Canon EOS D60")==0) {
	    Read_CF(pPhoto, length, recordOffset, dataOffset);
	 }

	 break;
      case RECORD_CAMERA_FIRMWARE:

	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;

	 assert(length < 256);
	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length); 

	 DEBUG_6( "%sFIRMWARE ID [%x] %d %x FIrmware [%s]\n",prefix, type, length, length, buffer);

	 break;
      case RECORD_EXIF_EXPOSURE:
	 DEBUG_4( "%sEXIF EXPOSURE DATA [%x] Length [%d]\n",prefix, type, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;
	 {
	    photo_exposure_EXIF_ciff photoExposureData;

	    ciff_Read_Record_Data(pPhoto, &photoExposureData, recordOffset, length); 

#define MOVE_INT2(a,b) pPhoto->a = Convert_Number_2_bytes(photoExposureData.b, pPhoto->byteOrder)
	    pPhoto->macroMode= Convert_Number_2_bytes(photoExposureData.macroMode, pPhoto->byteOrder);
	    pPhoto->selfTimer= Convert_Number_2_bytes(photoExposureData.selfTimer, pPhoto->byteOrder);
	    pPhoto->imageQuality= Convert_Number_2_bytes(photoExposureData.quality, pPhoto->byteOrder);
	    pPhoto->flashMode= Convert_Number_2_bytes(photoExposureData.flashMode, pPhoto->byteOrder);
	    pPhoto->continuousDrive = Convert_Number_2_bytes(photoExposureData.continuousDrive, pPhoto->byteOrder);

	    MOVE_INT2(focusingMode, focusingMode);

	    pPhoto->imageSizeRelative = Convert_Number_2_bytes(photoExposureData.imageSize, pPhoto->byteOrder);
	    pPhoto->easyProgramMode = Convert_Number_2_bytes(photoExposureData.easyShootingMode, pPhoto->byteOrder);
	    pPhoto->digitalZoom  = Convert_Number_2_bytes(photoExposureData.digitalZoom, pPhoto->byteOrder);
	    MOVE_INT2(imageContrast, imageContrast);
	    MOVE_INT2(imageSaturation, imageSaturation);
	    MOVE_INT2(imageSharpness, imageSharpness);
	    /*	    THis value is not reliable
	    MOVE_INT2(imageISO, imageISO);
	    */
	    MOVE_INT2(meteringMode, meteringMode);
	    MOVE_INT2(focusingType, focusingMode);
	    MOVE_INT2(autoFocusPointSelected, autoFocusPointSelected);
	    MOVE_INT2(exposureMode, exposureMode);
	    MOVE_INT2(lensFocalLengthMin, lensFocalLengthMin);
	    MOVE_INT2(lensFocalLengthMax, lensFocalLengthMax);
	    MOVE_INT2(lensFocalLengthUnits, lensFocalLengthUnits);

	    MOVE_INT2(flashInfo1, flashInfo1);
	    MOVE_INT2(flashInfo2, flashInfo2);
	    MOVE_INT2(flashActivity, flashActivity);
	    MOVE_INT2(flashDetails, flashDetails);


#undef MOVE_INT2

	    DEBUG_5("                             macro mode %d selftimelength %d quality %d FLash mode %d\n",
		    pPhoto->macroMode,
		    pPhoto->selfTimer,
		    pPhoto->imageQuality,
		    pPhoto->flashMode
		    );
	    DEBUG_5("                             continuos drive %d focus mode %d image size %d Program mode %d\n",
		    pPhoto->continuousDrive,
		    pPhoto->focusingMode,
		    pPhoto->imageSizeRelative,
		    pPhoto->easyProgramMode
		    );
	    DEBUG_5("                             digital zoom %d image conrast %d image saturation %d image sharpness %d\n",
		    pPhoto->digitalZoom,
		    pPhoto->imageContrast,
		    pPhoto->imageSaturation,
		    pPhoto->imageSharpness
		    );
	    DEBUG_5("                             ISO %d Metering mode %d focusing type %d Auto focus point selected %x\n",
		    pPhoto->imageISO,
		    pPhoto->meteringMode,
		    pPhoto->focusingType,
		    pPhoto->autoFocusPointSelected
		    );
	    DEBUG_5("                             Exposure Mode %d Focal Length Min %d Max %d Units per mm %d\n",
		    pPhoto->exposureMode,
		    pPhoto->lensFocalLengthMin,
		    pPhoto->lensFocalLengthMax,
		    pPhoto->lensFocalLengthUnits
		    );
	    DEBUG_8("                             flash info1 0x%x %d flash info 2 0x%x %d flashActivity 0x%x %d flashDetails %x\n",
		    pPhoto->flashInfo1, pPhoto->flashInfo1,
		    pPhoto->flashInfo2, pPhoto->flashInfo2,
		    pPhoto->flashActivity, pPhoto->flashActivity,
		    pPhoto->flashDetails
		    );
	    /*
	    DEBUG_6("                             UNKNOWNS u1 %u u2 %u u3 %u u4 %u u5 %u\n",
		    photoExposureData.unknown1,
		    photoExposureData.unknown2,
		    photoExposureData.unknown3,
		    photoExposureData.unknown4,
		    photoExposureData.unknown5
		    );

	    */
	    
	    if (debug)
	       Print_Bytes((char*)&photoExposureData, length);

	    if (debug) 
	       Print_Int2(&photoExposureData, length, pPhoto->byteOrder);
	 }
	 break;
      case 0x0001: /* D30 */
      case 0x0032: /* D30  Canon Color Info1*/
      case 0x0036: /* this record is identical in every photo I take*/
	 debug = 0;
	 DEBUG_4( "%s Color info? [%x] Length [%d]\n",prefix, type, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;
	 {
	    char buffer[8192];
	    assert(length <= 8192);
	    ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length); 

	    if (debug)
	       Print_Bytes(buffer, length);
	    /*
	    Print_Int2(buffer, length, pPhoto->byteOrder);
	    */

	 }
	 debug = 0;
	 break;
      case 0x10c2:
	 /* NO idea what this record is, but it is the same in all the photos */
	 DEBUG_1("Printing\n");
	 DEBUG_4( "%sUnknown [%x] Length [%d]\n",prefix, type, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;
	 if (debug)
	    Print_Bytes(buffer, length);
	 break;
      case 0x102c: /* D30 only CanonColorInfo2 */
      case 0x102d:
      case 0x107f: /* G2 only */
      case 0x10a8:
      case 0x10a9: /* Whitebalance Table */
      case 0x10aa:
      case 0x10c0:
      case 0x10c1:
      case 0x1093: /* d300 only */
      case 0x1038: /* d300 only   CanonPictureInfo */
      case 0x10ad: /* d300 only */
      case 0x10ae: /* d300 only */
      case 0x10af: /* d300 only */
      case 0x10b4: /* d300 only  Color Space --- 1 sRGB, 2 Adobe RGV, 0xfff uncalibrated */
      case 0x10b5: /* d300 only */
      case 0x1039: /* d300 only */
      case 0x183b: /* d300 only */
      case 0x1833:
      case 0x1030: /* D30 only */
      case 0x1835: /* Decoder table */
	 DEBUG_4( "%s*Record LONGS [%x] Length [%d]\n",prefix, type, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;
	 {
	    char buffer[4086];
	    assert(length < 4086);
	    ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length); 

	    if (debug) Print_Bytes(buffer, length);

	 }
	 break;

      case 0x2008: /* G2 only */

	 DEBUG_4( "%s*Record LONGS [%x] Length [%d]\n",prefix, type, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;


	 /*
	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length); 
	 Print_Bytes(buffer, length);
	 */
	 /*
	 for (j=0;j<length/4;j++){
	    dummyLong = Read_Int_4(pPhoto->file, pPhoto->byteOrder);
	    DEBUG_4( "%s    %2d Long %8lx %10lu\n", prefix, j, dummyLong, dummyLong);
	 }
	 */
	 break;
      case RECORD_SENSOR_DATA:
	 DEBUG_5( "%sSENSOR [%x] %d %x\n",prefix, type, length, length);
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;

	 assert(length < 256);
	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length); 
	 /*
	 Print_Int2(buffer, length, pPhoto->byteOrder);
	 */

	 break;

      case RECORD_IMAGE_DATE:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;

	 {
	    date_ciff dateRecord;
	    
	    long unsigned dateTT;

	    assert(length == sizeof(dateRecord));

	    ciff_Read_Record_Data(pPhoto,&dateRecord, recordOffset, length); 

	    dateTT = Convert_Number_4_bytes(dateRecord.dateTT, pPhoto->byteOrder);

	    DEBUG_7( "%sDATE [%x] %d %x TT [%lu] %s",prefix, type, length, length, dateTT, ctime(&dateTT));
	    pPhoto->imageDate = dateTT;
	 }

	 break;
      case RECORD_IMAGE_DESCRIPTION:	 

	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;


	 assert(length <= 256);
	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length); 

	 DEBUG_5( "%sIMAGE DESCRIPTION [%x] Length %d Value [%s]\n",prefix, type, length, buffer);

	 /*
	 for (j=0;j<length/4;j++){
	    dummyLong = Read_Int_4(pPhoto->file, pPhoto->byteOrder);
	    DEBUG_5( "%s    %2d Long %8lx %10lu\n", prefix, j, dummyLong, dummyLong);
	 }
	 */
	 break;

      case RECORD_IMAGE_DETAILS:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;
	 
	 {
	    image_description_ciff imageDescription;

	    assert(length == sizeof(image_description_ciff));
	    ciff_Read_Record_Data(pPhoto, &imageDescription, recordOffset, length);
	    
	    pPhoto->imageXSize = Convert_Number_4_bytes(imageDescription.xSize, pPhoto->byteOrder);
	    pPhoto->imageYSize = Convert_Number_4_bytes(imageDescription.ySize, pPhoto->byteOrder);
	    
	    DEBUG_7( "%sIMAGE DESCRIPTION [%x] %d %x X [%d] Y [%d] \n",
		    prefix, type, length, length, pPhoto->imageXSize, pPhoto->imageYSize);

	    /*
	      Print_Bytes(imageDescription.unknown1, sizeof(imageDescription.unknown1));
	    */
	 }


	 break;


      case RECORD_IMAGE_NAME: /* IMAGE_NAME  */
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset;

	 assert(length < 256);

	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);

	 DEBUG_6( "%sIMAGE NAME [%x] %d %d [%s]\n",prefix, type, length, length, buffer);

	 break;



      case RECORD_FILENAME:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) +  dataOffset;
	 assert(length < 256);
	 
	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);
	 strcpy(pPhoto->imageName, buffer);
	 DEBUG_6( "%sFilename [%x] %d %x [%s]\n",prefix, type, length, length, pPhoto->imageName);

	 break;

      case RECORD_OWNER_NAME:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) +  dataOffset;
	 assert(length < 256);
	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);

	 DEBUG_6( "%sOWNER NAME [%x] %d %x [%s]\n",prefix, type, length, length, buffer);

	 break;


      case RECORD_THM_FILENAME:

	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) +  dataOffset;
	 ciff_Read_Record_Data(pPhoto, buffer, recordOffset, length);
	 assert(length < 256);
	 DEBUG_6( "%sTHM filename [%x] %d %u [%s]\n",prefix, type, length, length, buffer);



	 break;
      case 0x5028:

	 DEBUG_5( "%s*XXXXXXXXXXX[%x] %d %d\n",prefix, type, length, length);
	 fread(dummy, 8, 1, pPhoto->file);
	 if (debug) Print_Bytes(dummy, 8);
	 break;

      case RECORD_FOCAL_LENGTH: /* Focal length */
	 {
	    focal_length_ciff focalLengthRaw;
	    fread(&focalLengthRaw, 8, 1, pPhoto->file);
	    pPhoto->imageFocalLength = Convert_Number_2_bytes(focalLengthRaw.focalLength, pPhoto->byteOrder);
	    DEBUG_5( "%sFOCAL Length[%x] %d FOCAL LENGTH %dmm\n",prefix, type, length, pPhoto->imageFocalLength);
	    if (debug) Print_Int2((char*)&focalLengthRaw, sizeof(focalLengthRaw),pPhoto->byteOrder);
	 }
	 break;

      case RECORD_PHOTO_SERIALNUMBER:
	 {
	    long serialNumber;
	    unsigned long unknown;
	    
	    serialNumber = Read_Int_4(pPhoto->file, pPhoto->byteOrder);
	    unknown = Read_Int_4(pPhoto->file, pPhoto->byteOrder);

	    DEBUG_6( "%sSERIAL NUMBER[%x] %d Photo Serial Number %ld Unknown %lu\n",
		    prefix, type, length, serialNumber, unknown);

	    pPhoto->imageSerialNumber = serialNumber;

	    /*
	    Print_Bytes((char*)&unknown, 4);
	    Print_Int2(&unknown, 4, pPhoto->byteOrder);
	    */
	 }


	 break;

      case RECORD_CAMERA_NUMBER:
	 {
	    unsigned long cameraNumber;
	    unsigned long unknown;
	    
	    cameraNumber = Read_Int_4(pPhoto->file, pPhoto->byteOrder);
	    unknown = Read_Int_4(pPhoto->file, pPhoto->byteOrder);

	    DEBUG_6( "%sCAMERA NUMBER[%x] %d Number %lu Unknown %lu\n",
		    prefix, type, length, cameraNumber, unknown);

	    pPhoto->cameraNumber = cameraNumber;

	 }


	 break;
      case 0x5803:
      case 0x5804:
      case 0x580b:
      case 0x5817:
      case 0x5814:
      case 0x5834: /* Model number? */
      case 0x501c:
      case RECORD_UNKNOWN1:
      case RECORD_UNKNOWN2:
      case RECORD_UNKNOWN3:
      case 0x480d:

	 debug=0;

	 DEBUG_5( "%s*Record UNKNOWN[%x] %d %d\n",prefix, type, length, length);
	 fread(dummy, 8, 1, pPhoto->file);

	 if (debug) {
	    Print_Bytes(dummy, 8);
	    Print_Int2(dummy, 8, pPhoto->byteOrder);
	    Print_Int4(dummy, 8, pPhoto->byteOrder);
	 }
	 debug = 0;

	 break;

	 /* These are the recursive records: */
      case 0x2807: /* Camera object */
      case 0x2804: /* Image spec */
      case 0x3003: /* Measured info */
      case 0x3005:
      case 0x300a:
      case 0x3004: /* Camera spec? */
      case 0x300b:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) + dataOffset ;

	 /*
	   DEBUG_6( "%sRecursive record %d %x---------------------------------\n",prefix, recordOffset, recordOffset);
	 */

	 switch (type) {
	 case 0x2807: /* Camera object */
	    DEBUG_2( "%sCAMERA OBJECT\n", prefix);
	    break;
	 case 0x2804: 
	    DEBUG_2( "%sIMAGE SPEC\n", prefix);
	    break;
	 case 0x3003: /* Measured info */
	    DEBUG_2( "%sMEASURED INFO\n", prefix);
	    break;
	 case 0x3004: 
	    DEBUG_2( "%sCAMERA SPEC\n", prefix);
	    break;
	 case 0x3005:
	    break;
	 case 0x300a:
	    DEBUG_2( "%sIMAGE PROPERTIES\n", prefix);
	    break;
	 case 0x300b:
	    DEBUG_2( "%sRAW IMAGE PROPERTIES\n", prefix);
	    break;
	 }
	 saveFileLocation = ftell(pPhoto->file);
	 Read_Records_Recursively(pPhoto, recordOffset, length, level+1);
	 fseek(pPhoto->file, saveFileLocation, SEEK_SET);
	 break;
      default:
	 DEBUG_6( "REading record %5d Type [%5x] type [%5d] length %8d %8x\n",
		 i, type, type, length, length);
	 
	 {
	    char str[80];
	    sprintf(str, "Unsuported record %0x\n", type);
	    My_Die(str);
	 }
      }



	 

   }
   
}




/*
 * Most frequently we are not interested in scanning the whole ciff data, we only want the 
 * location of the embedded JPEG
 * 
 * So far I have only found this file in the first level set of records, so don't even try to be
 * recursive. If it is not there, it is not there.
 */

static
int Find_Jpeg_Location_Only(craw_photo_type *pPhoto, long dataOffset, 
			    long lengthData)
{
   int count = 0;
   int i;
   long recordsOffset;

   assert(dataOffset > 0);
   assert(lengthData > 0);

   /* read the last 4 bytes from the data block */
   fseek(pPhoto->file, dataOffset+lengthData - 4, SEEK_SET);

   recordsOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder);

   /*
     DEBUG_4( "%sOffset of record header %ld %lx\n", prefix, recordsOffset,recordsOffset);
   */

   /*   photo_record *pRecords; */
   
   /* First 2 bytes are the number of records */

   fseek(pPhoto->file, recordsOffset + dataOffset, SEEK_SET);

   /*
     DEBUG_3( "%sSeek at %ld %lx\n", prefix, recordsOffset+dataOffset,recordsOffset+dataOffset);
   */

   count = Read_Int_2(pPhoto->file, pPhoto->byteOrder);

   DEBUG_3( ">Records [%d] %x\n", count, count);

   /* let us make sure we don't have weird values */
   assert(count> 0 && count <100);

   /* Allocate the records */

   for (i=0;i<count;i++) {
      int type =-1;
      long int length=-1;
      long int recordOffset=-1;
      

      type = Read_Int_2(pPhoto->file, pPhoto->byteOrder);
      if (type & 0x4000) {
	 fprintf(ERROR_OUTPUT, "This should not be reached\n");
	 assert(0);
      } else {
	 length = Read_Int_4(pPhoto->file, pPhoto->byteOrder);
      }



      switch (type) {
      case RECORD_JPEG:
	 recordOffset = Read_Int_4(pPhoto->file, pPhoto->byteOrder) +  dataOffset;
	 
	 pPhoto->jpegOffset = recordOffset;
	 pPhoto->jpegLength = length;
	 
	 DEBUG_3( "Embedded jpeg image %ld %ld\n", recordOffset, length);
	 return 1;
      case RECORD_RAW:
	 /* We need to skip the next 4 bytes */
	 Read_Int_4(pPhoto->file, pPhoto->byteOrder);
	 
	 break;
      default:
	 fprintf(ERROR_OUTPUT, "REading record %5d Type [%5x] type [%5d] length %8ld %8lx\n",
		 i, type, type, length, length);
	 
	 My_Die("Unsuported record\n");
      }
   }
   fprintf(ERROR_OUTPUT, "Embedded jpeg image not found\n");
   return 0;
}


int ciff_Find_Jpeg_Location(craw_photo_type *pPhoto)
{
   /* THe first record is at the end of the file */
   int result;

   result = Find_Jpeg_Location_Only(pPhoto, pPhoto->headerLength, 
				    pPhoto->fileLength - pPhoto->headerLength
				    );
   
   return result;
   
}



int ciff_Read_Data(craw_photo_type *pPhoto)
{
   /* THe first record is at the end of the file */
   

   Read_Records_Recursively(pPhoto, pPhoto->headerLength, 
				   pPhoto->fileLength - pPhoto->headerLength, 
				   0);

   return 1;
}


